"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
The following module contains the implementation for the  stacked observation
"""
import gymnasium
import numpy as np
from ray.rllib.models import ModelCatalog
from ray.rllib.models.tf.misc import normc_initializer
from ray.rllib.models.tf.tf_modelv2 import TFModelV2
from ray.rllib.models.utils import get_activation_fn
from ray.rllib.policy.view_requirement import ViewRequirement
from ray.rllib.utils.framework import try_import_tf
from ray.rllib.utils.spaces.space_utils import flatten_space
from ray.rllib.utils.typing import Dict, List, TensorType

tf1, tf, tfv = try_import_tf()


class FrameStackingModel(TFModelV2):
    """The following class is a slight modification of the base Fully Connected Model within RLLIB. The model adds on
        the ability to do frame stacking. This is done within the model to (1) enable the HPARAM search over the setting
        for the number of frames - Not possible with environment wrappers (2) enable a more in line view of the
        framestacking inline with paper representation.

        Note: Unlike the environment wrapper which just concatenates the data into one large vector and sends through
        network. This implementation will process N identical obs (vectors) through all hidden layers until the last
        FC layer which inputs are flattened to ensure that we are only producing the expected number of output actions

        The architecture produced is located below:
            - Note: Parallel paths. These are only generated if the Value function is defined as not sharing network.
                    In the case that the network is shared only a single path would exist.
            - Note: All of the default capability of the default RLLIB model is maintained and may not be shown in
                    diagram. Further testing needed for other paths.

                       FC Layers 1-N
                                        |-----|
                            |---------->|Dense|-|
                            |           |-----| |-|        |-------|     |-----|
                            |             |-----| |------->|Flatten|---->| FC  |---|
                            |               |-----|        |-------|     |-----|   |
        N Obs          |--------|                                                  |        Model
        -------------> |        |                                                  |
        (opt) N Reward |        |                                                  |------->[FC Out   ] ------> Actions
        -------------> | Inputs |
        (opt) N Action |        |                                                  |------->[Value Out] ------> Values
        -------------> |--------|                                                  |
                            |           |-----|                                    |
                            |---------->|Dense|-|                                  |
                                        |-----| |-|        |-------|     |-----|   |
                                          |-----| |------->|Flatten|---->| FC  |---|
                                            |-----|        |-------|     |-----|

        |----------------------------------------- |
        | Note: Ensure final output is in          |
        | 1 X ACTION not 1 X Frame X ACTION        |
        |------------------------------------------|

        Inputs  = 1 X Frames X (Obs + Rewards + Actions)
        Outputs = 1 X Actions

        The following is a example summary of the model from tensor flow on the  base Single Environments
        Model: "functional_1"
        __________________________________________________________________________________________________
        Layer (type)                    Output Shape         Param #     Connected to
        ==================================================================================================
        observations (InputLayer)       [(None, 5, 77)]      0
        __________________________________________________________________________________________________
        fc_0 (Dense)                    (None, 5, 256)       19968       observations[0][0]
        __________________________________________________________________________________________________
        fc_value_0 (Dense)              (None, 5, 256)       19968       observations[0][0]
        __________________________________________________________________________________________________
        fc_flatten_1 (Flatten)          (None, 1280)         0           fc_0[0][0]
        __________________________________________________________________________________________________
        fc_value_flatten_1 (Flatten)    (None, 1280)         0           fc_value_0[0][0]
        __________________________________________________________________________________________________
        fc_1 (Dense)                    (None, 256)          327936      fc_flatten_1[0][0]
        __________________________________________________________________________________________________
        fc_value_1 (Dense)              (None, 256)          327936      fc_value_flatten_1[0][0]
        __________________________________________________________________________________________________
        fc_out (Dense)                  (None, 51)           13107       fc_1[0][0]
        __________________________________________________________________________________________________
        value_out (Dense)               (None, 1)            257         fc_value_1[0][0]
        ==================================================================================================
        Total params: 709,172
        Trainable params: 709,172
        Non-trainable params: 0
        __________________________________________________________________________________________________

    Arguments:
        TFModelV2: [description]
    """

    PREV_N_OBS = "prev_n_obs"
    PREV_N_REWARDS = "prev_n_rewards"
    PREV_N_ACTIONS = "prev_n_actions"

    def __init__(
        self,
        obs_space,
        action_space,
        num_outputs,
        model_config,
        name,
        post_fcnet_hiddens=None,
        num_frames: int = 4,
        include_actions: bool = True,
        include_rewards: bool = True,
    ):
        """Class constructor

        Arguments:
            obs_space (gymnasium.spaces.Space): Observation space of the target gymnasium
                env. This may have an `original_space` attribute that
                specifies how to unflatten the tensor into a ragged tensor.
            action_space (gymnasium.spaces.Space): Action space of the target gymnasium
                env.
            num_outputs (int): Number of output units of the model.
            model_config (ModelConfigDict): Config for the model, documented
                in ModelCatalog.
            name (str): Name (scope) for the model.

        This method should create any variables used by the model.

        Keyword Arguments:
            num_frames {int} -- The number of frames to stack (default: {4})
            include_actions {int} -- Whether or not to include actions as part of frame stacking (default: True)
            include_actions {int} -- Whether or not to include actions as part of frame stacking (default: True)

        Returns:
            [type] -- [description]
        """
        if post_fcnet_hiddens is None:
            post_fcnet_hiddens = []
        # Initializes a ModelV2 object.
        TFModelV2.__init__(self, obs_space, action_space, num_outputs, model_config, name)

        # This model specific items
        self.num_frames = num_frames

        # Base model items
        self.num_outputs = num_outputs

        # Read out the model configuration parameters passed by RLLIB. Note this is maintained to ensure
        # compatibility with existing setup
        free_log_std, hiddens, activation, no_final_linear, vf_share_layers = self.get_config_opts()

        # Generate free-floating bias variables for the second half of
        # the outputs.
        if free_log_std:
            assert num_outputs % 2 == 0, ("num_outputs must be divisible by two", num_outputs)
            num_outputs = num_outputs // 2
            self.log_std_var = tf.Variable([0.0] * num_outputs, dtype=tf.float32, name="log_std")

        # Create the input layers for the observations, actions, and rewards
        flattened_action_space = flatten_space(action_space)
        observations, actions, rewards = self.create_input_layers(obs_space, flattened_action_space)

        # Select the input layer configuration based on input arguments
        self.include_rewards = include_rewards
        self.include_actions = include_actions
        self.input_list, self.inputs = FrameStackingModel.select_input_layer_configuration(
            include_rewards, include_actions, observations, actions, rewards
        )

        # Create layers 0 to second-last.
        last_layer = self.create_dense_hidden_layers(hiddens, self.inputs, activation, "fc")

        # The action distribution outputs.
        logits_out, last_layer = self.create_last_fc_layer_output(
            no_final_linear, num_outputs, activation, last_layer, hiddens, post_fcnet_hiddens, obs_space
        )

        # Concat the log std vars to the end of the state-dependent means.
        if free_log_std and logits_out is not None:

            def tiled_log_std(x):
                return tf.tile(tf.expand_dims(self.log_std_var, 0), [tf.shape(x)[0], 1])

            log_std_out = tf.keras.layers.Lambda(tiled_log_std)(self.inputs)
            logits_out = tf.keras.layers.Concatenate(axis=1)([logits_out, log_std_out])

        last_vf_layer = self.build_vf_network(vf_share_layers, self.inputs, hiddens, post_fcnet_hiddens, activation)

        value_out = tf.keras.layers.Dense(1, name="value_out", activation=None, kernel_initializer=normc_initializer(0.01))(
            last_vf_layer if last_vf_layer is not None else last_layer
        )

        self.base_model = tf.keras.Model(self.input_list, [(logits_out if logits_out is not None else last_layer), value_out])
        # print(self.base_model.summary())

        self.register_view_requirements(num_frames, obs_space, flattened_action_space)

        self._value_out = None

    @staticmethod
    def select_input_layer_configuration(include_rewards, include_actions, observations, actions, rewards):
        """Sets up the input layer based on the configuration of the arguments to the model

        Arguments:
            include_rewards {bool} -- Flag to indicate that rewards should be part of frame stacking
            include_actions {bool} -- Flag to indicate that the actions should be part of frame stacking
            observations {Tensor} -- [description]
            actions {Tensor} -- [description]
            rewards {Tensor} -- [description]
        """
        # Last hidden layer output (before logits outputs).
        if include_actions and not include_rewards:
            input_list = [observations, actions]
            inputs = tf.keras.layers.Concatenate(axis=-1)(input_list)
        elif not include_actions and include_rewards:
            input_list = [observations, rewards]
            inputs = tf.keras.layers.Concatenate(axis=-1)(input_list)
        elif include_actions and include_rewards:
            input_list = [observations, actions, rewards]
            inputs = tf.keras.layers.Concatenate(axis=-1)(input_list)
        else:
            input_list = [observations]
            inputs = observations
        return input_list, inputs

    def create_input_layers(self, obs_space, action_space):
        """Creates the input layers for starting the graph

        Arguments:
            obs_space {gymnasium.Space} -- The input space - flattended
            action_space {gymnasium.Space} -- The input space - flattended

        Returns:
            tuple[tensor] -- The input layers for observations, rewards, actions
        """
        # (?, Number of Frames, 1)
        rewards = tf.keras.layers.Input(shape=(self.num_frames, 1), name="rewards")
        # (?, Number of Frames, len obs flatten)
        observations = tf.keras.layers.Input(shape=(self.num_frames, obs_space.shape[0]), name="observations")
        # (?, Number of Frames, len actions flatten)
        actions = tf.keras.layers.Input(shape=(self.num_frames, len(action_space)), name="actions")
        return observations, actions, rewards

    def create_last_fc_layer_output(self, no_final_linear, num_outputs, activation, last_layer, hiddens, post_fcnet_hiddens, obs_space):
        """[summary]

        Arguments:
            no_final_linear: [description]
            num_outputs: [description]
            activation: [description]
            last_layer: [description]
            hiddens: [description]
            obs_space: [description]

        Returns:
            [type] -- [description]
        """
        # The action distribution outputs.
        logits_out = None
        # The last layer is adjusted to be of size num_outputs, but it's a
        # layer with activation.
        if no_final_linear and num_outputs:
            logits_out = tf.keras.layers.Dense(
                num_outputs, name="fc_out", activation=activation, kernel_initializer=normc_initializer(1.0)
            )(last_layer)

        # Finish the layers with the provided sizes (`hiddens`), plus -
        # iff num_outputs > 0 - a last linear layer of size num_outputs.
        else:
            if len(hiddens) > 0:
                last_layer = FrameStackingModel.flatten_plus_dense(
                    hiddens, post_fcnet_hiddens, last_layer, activation, "fc", len(hiddens) - 1
                )

            if num_outputs:
                logits_out = tf.keras.layers.Dense(num_outputs, name="fc_out", activation=None, kernel_initializer=normc_initializer(0.01))(
                    last_layer
                )
            # Adjust num_outputs to be the number of nodes in the last layer.
            else:
                self.num_outputs = ([int(np.prod(obs_space.shape))] + hiddens[-1:])[-1]
        return logits_out, last_layer

    def build_vf_network(self, vf_share_layers, inputs, hiddens, flatten_plus_dense, activation):
        """Creates the value function network if configured in model config

        Arguments:
            vf_share_layers: [description]
            inputs: [description]
            hiddens: [description]
            activation: [description]

        Returns:
            [type] -- [description]
        """
        last_vf_layer = None
        if not vf_share_layers:
            # Build a parallel set of hidden layers for the value net.
            value_function_prefix = "fc_value"
            last_vf_layer = self.create_dense_hidden_layers(hiddens, inputs, activation, value_function_prefix)
            last_vf_layer = self.flatten_plus_dense(
                hiddens, flatten_plus_dense, last_vf_layer, activation, value_function_prefix, len(hiddens) - 1
            )
        return last_vf_layer

    def register_view_requirements(self, num_frames: int, obs_space, flattened_action_space):
        """Sets up the view requirements for the forward pass call

        Arguments:
            num_frames {int} -- The number of frames to stack
            obs_space: The observation space definition
            flattened_action_space: flattened action space
        """
        self.view_requirements[FrameStackingModel.PREV_N_OBS] = ViewRequirement(
            data_col="obs", shift=f"-{num_frames - 1}:0", space=obs_space
        )
        if self.include_rewards:
            self.view_requirements[FrameStackingModel.PREV_N_REWARDS] = ViewRequirement(data_col="rewards", shift=f"-{self.num_frames}:-1")
        if self.include_actions:
            self.view_requirements[FrameStackingModel.PREV_N_ACTIONS] = ViewRequirement(
                data_col="actions",
                shift=f"-{self.num_frames}:-1",
                space=gymnasium.spaces.box.Box(low=-np.inf, high=np.inf, shape=(len(flattened_action_space),), dtype=np.int64),
            )

    def forward(
        self,
        input_dict: Dict[str, TensorType],
        state: List[TensorType],
        seq_lens: TensorType,
    ) -> (TensorType, List[TensorType]):  # type: ignore
        """Call the model with the given input tensors and state.

        Any complex observations (dicts, tuples, etc.) will be unpacked by __call__ before being passed to forward(). To access
        the flattened observation tensor, refer to input_dict[“obs”].

        This method can be called any number of times. In eager execution, each call to forward() will eagerly evaluate the model.
        In symbolic execution, each call to forward creates a computation graph that operates over the variables of this model
        (i.e., shares weights).

        Custom models should override this instead of __call__.

        Arguments:
            input_dict (dict) - dictionary of input tensors, including “obs”, “obs_flat”, “prev_action”, “prev_reward”, “is_training”,
                                “eps_id”, “agent_id”, “infos”, and “t”.
            state (list) - list of state tensors with sizes matching those returned by get_initial_state + the batch dimension
            seq_lens (Tensor) - 1d tensor holding input sequence lengths

        Returns:
            The model output tensor of size [BATCH, num_outputs], and the new RNN state.
        """

        if self.include_actions and not self.include_rewards:
            model_out, self._value_out = self.base_model(
                [input_dict[FrameStackingModel.PREV_N_OBS], input_dict[FrameStackingModel.PREV_N_ACTIONS]]
            )
        elif not self.include_actions and self.include_rewards:
            model_out, self._value_out = self.base_model(
                [input_dict[FrameStackingModel.PREV_N_OBS], input_dict[FrameStackingModel.PREV_N_REWARDS]]
            )
        elif self.include_actions and self.include_rewards:
            model_out, self._value_out = self.base_model(
                [
                    input_dict[FrameStackingModel.PREV_N_OBS],
                    input_dict[FrameStackingModel.PREV_N_ACTIONS],
                    input_dict[FrameStackingModel.PREV_N_REWARDS],
                ]
            )
        else:
            model_out, self._value_out = self.base_model([input_dict[FrameStackingModel.PREV_N_OBS]])
        return model_out, state

    def value_function(self) -> TensorType:
        """Returns the value function output for the most recent forward pass.

        Note that a forward call has to be performed first, before this methods can return anything and thus
        that calling this method does not cause an extra forward pass through the network.

        Returns:
            value estimate tensor of shape [BATCH].
        """
        return tf.reshape(self._value_out, [-1])

    def get_config_opts(self):
        """Gets the configuration options utilizes by the frame stacking model

        Returns:
            Tuple -- configuration options for the models (Bool, list[int], function, Bool, Bool)
        """
        hiddens = self.model_config.get("fcnet_hiddens", []) + self.model_config.get("post_fcnet_hiddens", [])
        activation = self.model_config.get("fcnet_activation")
        if not self.model_config.get("fcnet_hiddens", []):
            activation = self.model_config.get("post_fcnet_activation")
        activation = get_activation_fn(activation)
        no_final_linear = self.model_config.get("no_final_linear")
        vf_share_layers = self.model_config.get("vf_share_layers")
        free_log_std = self.model_config.get("free_log_std")
        return free_log_std, hiddens, activation, no_final_linear, vf_share_layers

    @staticmethod
    def create_dense_hidden_layers(hiddens, layer, activation, prefix: str):
        """Creates the hidden dense layers

        Arguments:
            hiddens {List[int]} -- The list of hidden layers for the FC components
            layer {Tensor} -- [description] --- TODO Remove as not needed to pass in...
            activation {Function} -- [description]
            prefix {str} -- The string to use for the naming of the layer

        Returns:
            [type] -- [description]
        """
        for index, size in enumerate(hiddens[:-1]):
            dense_name = f"{prefix}_{index}"
            layer = tf.keras.layers.Dense(size, name=dense_name, activation=activation, kernel_initializer=normc_initializer(1.0))(layer)
        return layer

    @staticmethod
    def flatten_plus_dense(hiddens, post_fcnet_hiddens, layer, activation, prefix: str, index: int):
        """Creates the final/last dense layer with flatten to ensure the correct output size

        Arguments:
            hiddens {List[int]} -- List containing the size of each hidden layer
            layer {Tensor} -- [description]
            activation {function]} -- [description]
            prefix {str} -- The string to use when creating the layers
            index {int} -- The index of the layer to add the flatten on.

        Returns:
            Tensor -- The layer just created with flatten + FD
        """
        flatten_name = f"{prefix}_flatten_{index}"
        dense_name = f"{prefix}_{index}"
        layer = tf.keras.layers.Flatten(name=flatten_name)(layer)
        layer = tf.keras.layers.Dense(hiddens[index], name=dense_name, activation=activation, kernel_initializer=normc_initializer(1.0))(
            layer
        )
        for index_cat, size in enumerate(post_fcnet_hiddens):
            layer = tf.keras.layers.Dense(
                size, name=f"{dense_name}_cat_{index_cat}", activation=activation, kernel_initializer=normc_initializer(1.0)
            )(layer)
        return layer


ModelCatalog.register_custom_model("FrameStackingModel", FrameStackingModel)



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

Glue which transforms observation into a unit vector and allows custom normalization.

Author: Jamie Cunningham
"""

from collections import OrderedDict
from functools import cached_property

import gymnasium
import numpy as np

from corl.glues.base_wrapper import BaseWrapperGlue, BaseWrapperGlueValidator
from corl.glues.common.observe_sensor import ObserveSensor
from corl.libraries.property import BoxProp, DictProp
from corl.libraries.units import corl_quantity


class UnitVectorGlueValidator(BaseWrapperGlueValidator):
    """
    Validator for UnitVector glue
    """

    wrapped: ObserveSensor


class UnitVectorGlue(BaseWrapperGlue):
    """
    Transforms an observation into a unit vector.
    """

    class Fields:
        """
        field data
        """

        UNIT_VEC = "unit_vec"

    def __init__(self, **kwargs) -> None:
        self.config: UnitVectorGlueValidator
        super().__init__(**kwargs)

        self._uname = f"{self.glue().get_unique_name()}_UnitVector"

    @staticmethod
    def get_validator() -> type[UnitVectorGlueValidator]:
        return UnitVectorGlueValidator

    def get_unique_name(self):
        return self._uname

    @cached_property
    def observation_prop(self):
        tmp = self.glue().observation_space[self.config.wrapped.Fields.DIRECT_OBSERVATION].shape
        return DictProp(
            spaces={
                self.Fields.UNIT_VEC: BoxProp(
                    low=[-1.0] * tmp[0], high=[1.0] * tmp[0], shape=tmp, dtype=np.dtype("float32"), unit="dimensionless"
                )
            }
        )

    def get_observation(self, other_obs: OrderedDict, obs_space, obs_units) -> OrderedDict:
        direct_obs = self.config.wrapped.get_observation(other_obs, obs_space, obs_units)[self.config.wrapped.Fields.DIRECT_OBSERVATION]

        # Compute unit vector
        norm = np.linalg.norm(direct_obs.m)
        unit_vec = direct_obs.m / norm

        # Return observation
        d = OrderedDict()
        d[self.Fields.UNIT_VEC] = corl_quantity()(np.array(unit_vec, dtype=np.float32), "dimensionless")
        return d

    @cached_property
    def action_space(self) -> gymnasium.spaces.Space | None:
        """No Actions"""
        return None

    def apply_action(self, action, observation, action_space, obs_space, obs_units):  # noqa: PLR6301
        """No Actions"""
        return



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import typing

from pydantic import BaseModel, ConfigDict, field_validator

from corl.evaluation.metrics.metric import Metric, NonTerminalMetric
from corl.libraries.units import Quantity, corl_get_ureg


class TimedValue(BaseModel, NonTerminalMetric):
    """NonTerminal Metric that contains a metric which occurred at a specific time"""

    time: Quantity
    value: Metric
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("time")
    @classmethod
    def time_unit_must_be_time(cls, v: Quantity):
        """Enusure that the units given are of Time Dimension"""
        assert v.u == corl_get_ureg().get_unit("second")
        return v

    def __add__(self, rhs: typing.Any):
        raise NotImplementedError()

    def __truediv__(self, rhs: typing.Any):
        raise NotImplementedError()



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Provides AvailablePlatformTypes for the Open ML Gymnasium Simulator
"""
from __future__ import annotations

from corl.libraries.plugin_library import PluginLibrary
from corl.simulators.base_platform_type import BasePlatformType
from corl.simulators.gymnasium.gymnasium_simulator import GymnasiumSimulator


class GymnasiumAvailablePlatformTypeMain(BasePlatformType):
    """_summary_

    Args:
        PlatformType (_type_): _description_

    Returns:
        _type_: _description_
    """

    @classmethod
    def match_model(cls, config: dict) -> bool:
        """_summary_

        Args:
            config (dict): _description_

        Returns:
            bool: _description_
        """
        return True


PluginLibrary.add_platform_to_sim(GymnasiumAvailablePlatformTypeMain, GymnasiumSimulator)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import gc
import importlib
import logging
import os
import pathlib
import re
import socket
import sys
import typing
import warnings
from datetime import datetime

import git
import ray
from pydantic import AfterValidator, BaseModel, ConfigDict, ImportString, field_validator
from ray import tune
from ray.rllib.algorithms.callbacks import DefaultCallbacks, make_multi_callbacks
from ray.rllib.policy.policy import Policy
from ray.tune.registry import get_trainable_cls, register_env

import corl.experiments.rllib_utils.wrappers as rllib_wrappers
from corl.environment.base_multi_agent_env import BaseCorlMultiAgentEnv
from corl.environment.default_env_rllib_callbacks import EnvironmentDefaultCallbacks
from corl.environment.environment_wrappers import GroupedAgentsEnv
from corl.environment.multi_agent_env import ACT3MultiAgentEnv
from corl.episode_parameter_providers import EpisodeParameterProvider
from corl.episode_parameter_providers.remote import RemoteEpisodeParameterProvider
from corl.experiments.base_experiment import BaseExperiment, BaseExperimentValidator, ExperimentFileParse
from corl.experiments.rllib_utils.policy_mapping_functions import PolicyIsAgent
from corl.libraries.factory import Factory
from corl.libraries.file_path import CorlDirectoryPath
from corl.parsers.yaml_loader import apply_patches, load_file
from corl.policies.base_policy import BasePolicyValidator


class RllibPolicyMappingValidator(BaseModel):
    """validator of data used to generate policy mapping"""

    functor: ImportString = PolicyIsAgent
    config: dict = {}


class RllibExperimentValidator(BaseExperimentValidator):
    """
    ray_config: dictionary to be fed into ray init, validated by ray init call
    env_config: environment configuration, validated by environment class
    rllib_configs: a mapping of compute platforms to rllib configs, see apply_patches_rllib_configs
                    for information on the typing
    tune_config: kwarg arguments to be sent to tune for this experiment
    extra_callbacks: extra rllib callbacks that will be added to the callback list
    trial_creator_function: this function will overwrite the default trial string creator
                            and allow more fine tune trial name creators
    """

    ray_config: dict[str, typing.Any]
    env_config: dict[str, typing.Any]
    rllib_configs: dict[str, dict[str, typing.Any]]
    tune_config: dict[str, typing.Any]
    trainable_config: dict[str, typing.Any] | None = None
    hparam_search_class: ImportString | None = None
    hparam_search_config: dict[str, typing.Any] | None = None
    extra_callbacks: list[ImportString] | None = None
    extra_tune_callbacks: list[ImportString] | None = None
    trial_creator_function: ImportString | None = None
    policy_mapping: RllibPolicyMappingValidator = RllibPolicyMappingValidator()
    environment: str = "CorlMultiAgentEnv"
    policies: dict[str, str] = {}
    configuration_only: bool = False
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("rllib_configs", mode="before")
    @classmethod
    def apply_patches_rllib_configs(cls, v):
        """
        The dictionary of rllib configs may come in as a dictionary of
        lists of dictionaries, this function is responsible for collapsing
        the list down to a typing.Dict[str, typing.Dict[str, typing.Any]]
        instead of
        typing.Dict[str, typing.Union[typing.List[typing.Dict[str, typing.Any]], typing.Dict[str, typing.Any]]]

        Raises:
            RuntimeError: [description]

        Returns:
            [type] -- [description]
        """
        if not isinstance(v, dict):
            raise RuntimeError("rllib_configs are expected to be a dict of keys to different compute configs")
        rllib_configs = {}
        for key, value in v.items():
            if isinstance(value, list):
                warnings.warn("having this validator apply patches is deprecated, please use the !merge yml directive", DeprecationWarning)
                rllib_configs[key] = apply_patches(value)
            elif isinstance(value, dict):
                rllib_configs[key] = value
        return rllib_configs

    @field_validator("ray_config", "tune_config", "trainable_config", "env_config", mode="before")
    @classmethod
    def apply_patches_configs(cls, v):
        """
        reduces a field from
        typing.Union[typing.List[typing.Dict[str, typing.Any]], typing.Dict[str, typing.Any]]]
        to
        typing.Dict[str, typing.Any]

        by patching the first dictionary in the list with each patch afterwards

        Returns:
            [type] -- [description]
        """
        if isinstance(v, list):
            warnings.warn("having this validator apply patches is deprecated, please use the !merge yml directive", DeprecationWarning)
            # OLD METHOD:
            #     env_config: [!include ../environments/core_capture_v1.yml, *env_config_updates]
            #
            # N.B. in this case the environment command line argument is not used.
            v = apply_patches(v)
        return v


def path_exists(v):
    """Check that the checkpoint path exists"""

    if v is None:
        return v

    sub_file_checks = []
    for filename in ["rllib_checkpoint.json", "policy_state.pkl"]:
        tmp = v / filename
        sub_file_checks.append(tmp.exists())
    if not all(sub_file_checks):
        raise RuntimeError(
            "Loading a checkpoint from the policy config 'checkpoint_path' requires "
            "specifying a rllib policy folder that has both a 'rllib_checkpoint.json' and a 'policy_state.pkl' in it."
        )
    return v


class RllibPolicyValidator(BasePolicyValidator):
    """
    policy_class: callable policy class None will use default from trainer
    checkpoint_path: if this policy should attempt to load weights from a checkpoint
    train: should this policy be trained
    Arguments:
        BaseModel: [description]

    Raises:
        RuntimeError: [description]

    Returns:
        [type] -- [description]
    """

    config: dict[str, typing.Any] = {}
    policy_class: ImportString | None = None
    checkpoint_path: typing.Annotated[CorlDirectoryPath, AfterValidator(path_exists)] | None = None
    train: bool = True


class RllibExperiment(BaseExperiment):
    """
    The Rllib Experiment is an experiment for running
    multiagent configurable environments with patchable settings
    """

    def __init__(self, **kwargs) -> None:
        self.config: RllibExperimentValidator
        self._logger = logging.getLogger(RllibExperiment.__name__)
        super().__init__(**kwargs)
        self._register_envs()

    @classmethod
    def process_cli_args(cls, config: dict, cli_args: ExperimentFileParse):
        """Modifies/overrides config fields with command line arguments."""
        in_env_config = config["env_config"]

        # OLD METHOD:
        #     env_config: [!include ../environments/core_capture_v1.yml, *env_config_updates]
        #
        #   N.B. in this case the environment command line argument is not used.
        #
        # NEW METHOD
        #     env_config:
        #         default: [!include ../environments/core_capture_v1.yml, *env_config_updates]
        #         option1: [!include ../environments/core_capture_v1.yml, *env_config_updates]
        #         option2: [!include ../environments/core_capture_v1_mod.yml, *env_config_updates]
        if isinstance(in_env_config, dict):
            if cli_args.environment in in_env_config:
                config["env_config"] = config["env_config"][cli_args.environment]
            elif "simulator" in in_env_config and "episode_parameter_provider" in in_env_config:
                pass
            else:
                raise ValueError(
                    f"The selected env is not in the env_config dict --- Environment = {cli_args.environment} --- {in_env_config}"
                )

    @staticmethod
    def get_validator() -> type[RllibExperimentValidator]:
        return RllibExperimentValidator

    @property
    def get_policy_validator(self) -> type[RllibPolicyValidator]:
        """Return validator"""
        return RllibPolicyValidator

    def initialize_ray(self, compute_platform: str, debug: bool):
        """Initializes ray"""
        if compute_platform in ["ray"]:
            self._update_ray_config_for_ray_platform()

        if debug:
            self.config.ray_config["local_mode"] = True

        ray.init(**self.config.ray_config)

    def run_experiment(self, args: ExperimentFileParse) -> None:
        # Check for env key in env_config and if not present allow deprecated
        # path to still work --- else index and replace with selected env.
        rllib_config = self._select_rllib_config(args.compute_platform)

        self.initialize_ray(args.compute_platform, args.debug)

        self.config.env_config["agents"], self.config.env_config["agent_platforms"] = self.create_agents(
            args.platform_config, args.agent_config
        )

        if args.output:
            self.config.env_config["output_path"] = str(args.output)
            self.config.tune_config["storage_path"] = str(args.output)

        if args.name:
            self.config.env_config["TrialName"] = args.name

        self._add_trial_creator()

        if not self.config.ray_config["local_mode"]:
            self.config.env_config["episode_parameter_provider"] = RemoteEpisodeParameterProvider.wrap_epp_factory(
                Factory(**self.config.env_config["episode_parameter_provider"]),
                actor_name=BaseCorlMultiAgentEnv.episode_parameter_provider_name,
            )

            for agent_name, agent_configs in self.config.env_config["agents"].items():
                agent_configs.class_config.config["episode_parameter_provider"] = RemoteEpisodeParameterProvider.wrap_epp_factory(
                    Factory(**agent_configs.class_config.config["episode_parameter_provider"]), agent_name
                )
        tmp_env: ACT3MultiAgentEnv = self.create_environment()
        self.config.env_config["epp_registry"] = tmp_env.config.epp_registry
        policies, train_policies, available_policies = self.create_policies(tmp_env)

        self._update_rllib_config(rllib_config, train_policies, policies, args, available_policies)
        self._update_tune_config()

        self._enable_episode_parameter_provider_checkpointing()

        search_class = None
        if self.config.hparam_search_class is not None:
            if self.config.hparam_search_config is not None:
                search_class = self.config.hparam_search_class(**self.config.hparam_search_config)
            else:
                search_class = self.config.hparam_search_class()
            search_class.add_algorithm_hparams(rllib_config, self.config.tune_config)

        self.on_configuration_end(args, rllib_config)

        if self.config.configuration_only:
            return
        # this environment may have initialized something on worker process, so just be to sure it is not going to collide
        # with anything
        del tmp_env
        gc.collect()

        tune.run(
            config=rllib_config,
            # progress_reporter=(),
            **self.config.tune_config,
        )

    def create_environment(self) -> typing.Any:
        """Return environment from rllib environment creators"""
        return rllib_wrappers.get_rllib_environment_creator(self.config.environment)(self.config.env_config)

    def create_policies(self, environment):
        """Return various polices used in experiments"""
        env_policies = {
            agent_name: self.get_policy_validator(**agent_parse.policy_config)
            for agent_name, agent_parse in self.config.env_config["agents"].items()
        }
        experiment_policies = {name: self.get_policy_validator(**load_file(config)) for name, config in self.config.policies.items()}
        available_policies = experiment_policies | env_policies

        policies = {
            policy_name: (
                available_policies[policy_name].policy_class,
                policy_obs,
                environment.action_space[policy_name],
                available_policies[policy_name].config,
            )
            for policy_name, policy_obs in environment.observation_space.spaces.items()
            if available_policies[policy_name]
        }

        train_policies = [policy_name for policy_name in policies if available_policies[policy_name].train]
        return policies, train_policies, available_policies

    def on_configuration_end(self, args: ExperimentFileParse, rllib_config: dict[str, typing.Any]) -> None:
        """Subclass hook for additional configuration after the end of the standard configuration."""

    def _update_rllib_config(self, rllib_config, train_policies, policies, args: ExperimentFileParse, available_policies) -> None:
        """
        Update several rllib config fields
        tmp_ac: agent_classes from temporary environment to get obs/action space, these should be
                considered read only, and will be destroyed when training begins
        """

        rllib_config["multiagent"] = {
            "policies": policies,
            "policy_mapping_fn": self.config.policy_mapping.functor(**self.config.policy_mapping.config),
            "policies_to_train": train_policies,
        }

        rllib_config["env"] = self.config.environment
        callback_list: list[type[DefaultCallbacks]] = [self.get_callbacks()]
        if checkpoint_weight_files := {
            agent_name: agent_config.checkpoint_path
            for agent_name, agent_config in available_policies.items()
            if agent_config.checkpoint_path
        }:

            class WeightLoaderCallback(DefaultCallbacks):
                """
                callback that will load the preexisting weights
                into policies
                """

                def on_algorithm_init(self, *, algorithm, **kwargs):  # noqa: PLR6301
                    agent_weight_dict = {}
                    for agent_name, file_path in checkpoint_weight_files.items():
                        try:
                            git_repo = git.Repo(file_path, search_parent_directories=True)
                        except git.InvalidGitRepositoryError:
                            print(f"Checkpoint hash for {agent_name} at {file_path} is unknown")
                        else:
                            git_hash = git_repo.head.object.hexsha
                            print(f"Checkpoint hash for {agent_name} at {file_path}: {git_hash}")
                        restored = Policy.from_checkpoint(file_path)
                        agent_weight_dict[agent_name] = restored.get_weights()
                    algorithm.set_weights(agent_weight_dict)
                    algorithm.workers.sync_weights()

            callback_list.append(WeightLoaderCallback)
        if self.config.extra_callbacks:
            callback_list.extend(self.config.extra_callbacks)
        rllib_config["callbacks"] = make_multi_callbacks(callback_list)
        rllib_config["env_config"] = self.config.env_config
        now = datetime.now()
        rllib_config["env_config"]["output_date_string"] = f"{now.strftime('%Y%m%d_%H%M%S')}_{socket.gethostname()}"
        rllib_config["create_env_on_driver"] = True

        self._add_git_hashes_to_config(rllib_config)

        if args.debug:
            rllib_config["num_workers"] = 0

    def _update_tune_config(self) -> None:
        """
        Update tune config with extra callbacks
        """
        if self.config.extra_tune_callbacks:
            self.config.tune_config["callbacks"] = [callback() for callback in self.config.extra_tune_callbacks]

    def _add_git_hashes_to_config(self, rllib_config) -> None:
        """adds git hashes (or package version information if git information
        is unavailable) of key modules to rllib_config["env_config"]["git_hash"].
        Key modules are the following:
          - corl,
          - whatever cwd is set to at the time of the function call
            (notionally /opt/project / -rl-agents)
          - any other modules listed in rllib_config["env_config"]["plugin_paths"]

        This information is not actually used by BaseCorlMultiAgentEnv;
        however, putting it in the env_config means that this
        information is saved to the params.pkl and thus is available
        for later inspection while seeking to understand the
        performance of a trained model.
        """
        try:
            # pattern used below to find root repository paths
            repo_pattern = r"(?P<repopath>.*)\/__init__.py"
            rp = re.compile(repo_pattern)

            corl_pattern = r"corl.*"
            cp0 = re.compile(corl_pattern)
            rllib_config["env_config"]["git_hash"] = {}

            # store hash on cwd
            cwd = os.getcwd()
            try:
                githash = git.Repo(cwd, search_parent_directories=True).head.object.hexsha
                rllib_config["env_config"]["git_hash"]["cwd"] = githash
                self._logger.info(f"cwd hash: {githash}")
            except git.InvalidGitRepositoryError:
                self._logger.warning("cwd is not a git repo\n")

            # go ahead and strip out corl related things from plugin_path
            plugpath = []
            for item in rllib_config["env_config"]["plugin_paths"]:
                match0 = cp0.match(item)
                if match0 is None:
                    plugpath.append(item)

            plugpath.append("corl")

            # add git hashes to env_config dictionary
            for module0 in plugpath:
                env_hash_key = module0
                module1 = importlib.import_module(module0)
                modulefile = module1.__file__
                if modulefile is not None:
                    match0 = rp.match(modulefile)
                    if match0 is not None:
                        repo_path = match0.group("repopath")
                        try:
                            githash = git.Repo(repo_path, search_parent_directories=True).head.object.hexsha
                            rllib_config["env_config"]["git_hash"][env_hash_key] = githash
                            self._logger.info(f"{module0} hash: {githash}")
                        except git.InvalidGitRepositoryError:
                            # possibly installed in image but not a git repo
                            # look for version number
                            if hasattr(module1, "version") and hasattr(module1.version, "__version__"):
                                githash = module1.version.__version__
                                rllib_config["env_config"]["git_hash"][env_hash_key] = githash
                                self._logger.info(f"{module0} hash: {githash}")
                            else:
                                self._logger.warning(f"module: {module0}, repopath: {repo_path} is invalid git repo\n")
                                sys.stderr.write(f"module: {module0}, repopath: {repo_path} is invalid git repo\n")
        except ValueError:
            warnings.warn("Unable to add the github hash to experiment!!!")

    def get_callbacks(self) -> type[EnvironmentDefaultCallbacks]:  # noqa: PLR6301
        """Get the environment callbacks"""
        return EnvironmentDefaultCallbacks

    def _register_envs(self):  # noqa: PLR6301
        register_env("CorlMultiAgentEnv", lambda config: ACT3MultiAgentEnv(config))
        register_env("CorlGroupedAgentEnv", lambda config: GroupedAgentsEnv(config))

    def _select_rllib_config(self, platform: str | None) -> dict[str, typing.Any]:
        """Extract the rllib config for the proper computational platform

        Parameters
        ----------
        platform : typing.Optional[str]
            Specification of the computational platform to use, such as "local", "hpc", etc.  This must be present in the rllib_configs.
            If None, the rllib_configs must only have a single entry.

        Returns
        -------
        typing.Dict[str, typing.Any]
            Rllib configuration for the desired computational platform.

        Raises
        ------
        RuntimeError
            The requested computational platform does not exist or None was used when multiple platforms were defined.
        """
        if platform is not None:
            return self.config.rllib_configs[platform]

        if len(self.config.rllib_configs) == 1:
            return self.config.rllib_configs[next(iter(self.config.rllib_configs))]

        raise RuntimeError(f'Invalid rllib_config for platform "{platform}"')

    def _update_ray_config_for_ray_platform(self) -> None:
        """Update the ray configuration for ray platforms"""
        self.config.ray_config["address"] = "auto"
        self.config.ray_config["log_to_driver"] = False

    def _enable_episode_parameter_provider_checkpointing(self) -> None:
        base_trainer = self.config.tune_config["run_or_experiment"]

        trainer_class = get_trainable_cls(base_trainer)

        def save_checkpoint(cls, checkpoint_path: str) -> None:
            """
            adds additional checkpoint saving functionality
            by also saving any episode parameter providers
            currently running
            """
            # mypy complains because the first argument to super isn't the current
            # `self` but this is actually OK since we in fact don't want super to
            # ascend the MRO of `self` but rather of `trainer_class`
            tmp = super(trainer_class, cls).save_checkpoint(checkpoint_path)  # type: ignore

            checkpoint_folder = pathlib.Path(checkpoint_path)

            # Environment
            epp_name = ACT3MultiAgentEnv.episode_parameter_provider_name
            env = cls.workers.local_worker().env
            epp: EpisodeParameterProvider = env.config.epp
            epp.save_checkpoint(checkpoint_folder / epp_name)

            # Agents
            for agent_name, agent_configs in env.agent_dict.items():
                epp = agent_configs.config.epp
                epp.save_checkpoint(checkpoint_folder / agent_name)

            return tmp

        def load_checkpoint(cls, checkpoint_path: str) -> None:
            """
            adds additional checkpoint loading functionality
            by also loading any episode parameter providers
            with the checkpoint
            """
            # mypy complains because the first argument to super isn't the current
            # `self` but this is actually OK since we in fact don't want super to
            # ascend the MRO of `self` but rather of `trainer_class`
            super(trainer_class, cls).load_checkpoint(checkpoint_path)  # type: ignore

            checkpoint_folder = pathlib.Path(checkpoint_path)

            # Environment
            epp_name = ACT3MultiAgentEnv.episode_parameter_provider_name
            env = cls.workers.local_worker().env
            epp: EpisodeParameterProvider = env.config.epp
            epp.load_checkpoint(checkpoint_folder / epp_name)

            # Agents
            for agent_name, agent_configs in env.agent_dict.items():
                epp = agent_configs.config.epp
                epp.load_checkpoint(checkpoint_folder / agent_name)

        # NOTE: this adds the EPP checkpointing while preserving the trainable name
        trainer_class.save_checkpoint = save_checkpoint
        trainer_class.load_checkpoint = load_checkpoint

        self.config.tune_config["run_or_experiment"] = trainer_class

    def _add_trial_creator(self):
        """Updates the trial name based on the HPC Job Number and the trial name in the configuration"""
        if "trial_name_creator" not in self.config.tune_config:
            if self.config.trial_creator_function is None:
                trial_prefix = os.environ.get("PBS_JOBID", os.environ.get("TRIAL_NAME_PREFIX", ""))
                trial_name = ""

                if "TrialName" in self.config.env_config:
                    trial_name = "-" + self.config.env_config["TrialName"] if trial_prefix else self.config.env_config["TrialName"]

                def trial_name_prefix(trial):
                    """
                    Args:
                        trial (Trial): A generated trial object.

                    Returns:
                        trial_name (str): String representation of Trial prefixed
                            by the contents of the environment variable:
                            PBS_JOBID or TRIAL_NAME_PREFIX
                            followed by the TrialName element of the environment configuration
                    """
                    return f"{trial_prefix}{trial_name}-{trial}"

                self.config.tune_config["trial_name_creator"] = trial_name_prefix
            else:
                self.config.tune_config["trial_name_creator"] = self.config.trial_creator_function



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from typing import Any

from pydantic import BaseModel, ImportString, ValidationInfo


def _construct(type_, info, **kwargs):
    if info and info.context:
        if issubclass(type_, BaseModel):
            return type_.model_validate(kwargs, context=info.context)

        try:
            validated_args = type_.get_validator().model_validate(kwargs, context=info.context)
            return type_(**dict(validated_args))
        except AttributeError:
            pass

    return type_(**kwargs)


class Factory(BaseModel):
    """Factory class to allow subclass creation with pydantic

    TODO: Create a simple motivating example
    """

    type: ImportString  # noqa: A003
    wrapped: dict[str, "Factory"] = {}
    config: dict[str, Any] = {}

    def build(self, info: ValidationInfo | None = None, **kwargs):
        """Build the object contained within this factory."""

        ctor_kwargs = {**self.config, **kwargs}
        if self.wrapped:
            ctor_kwargs["wrapped"] = {k: v.build(**kwargs) for k, v in self.wrapped.items()}

        return _construct(self.type, info, **ctor_kwargs)

    @classmethod
    def resolve_factory(cls, v, info):
        """Validator for converting a factory into the built object.

        Usage in a pydantic model:
        resolve_factory = field_validator('name')(Factory.resolve_factory)
        """
        try:
            v["type"]
        except (TypeError, KeyError):
            # Not something that should be built with the factory
            return v
        else:
            factory = cls(**v)
            return factory.build(info=info)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
EpisodeParameterProvider that reads data from a tabular structure.
"""
import copy
import typing

import pandas as pd
from pydantic import field_validator, model_validator

from corl.episode_parameter_providers import EpisodeParameterProvider, EpisodeParameterProviderValidator, ParameterModel, Randomness
from corl.libraries.parameters import OverridableParameterWrapper, Parameter


class TabularValidator(EpisodeParameterProviderValidator):
    """Validation model for the inputs of TabularParameterProvider"""

    separator: str = "/"
    filename: str | None = None
    read_csv_kwargs: dict[str, typing.Any] = {}
    data: pd.DataFrame

    @field_validator("separator")
    @classmethod
    def separator_validator(cls, v):
        """Validate that the separator is a single character"""
        if len(v) != 1:
            raise ValueError("Separator must be single character")
        return v

    @model_validator(mode="before")
    def data_validator(cls, values):
        """Validate the data field"""
        if values.get("data") is None:
            if values.get("filename") is None:
                raise ValueError("Either data or filename must be provided")
            values["data"] = pd.read_csv(values["filename"], **values["read_csv_kwargs"])
        return values

    @field_validator("data")
    def column_validator(cls, v, info):
        """Validate the columns in the data field"""
        parameters_keys = {info.data["separator"].join(x) for x in info.data["parameters"]}
        data_columns = set(v.columns)
        if parameters_keys != data_columns:
            left_side = parameters_keys - data_columns
            right_side = data_columns - parameters_keys
            raise RuntimeError(
                "TabularParameterProvider data table columns do not match the parameters keys, "
                f"parameters currently expected are {parameters_keys}\n"
                f"parameters not covered by config are {left_side}\n"
                f"parameters that are extra in config are {right_side}"
            )
        # assert parameters_keys == data_columns, "TabularParameterProvider data table columns do not match the parameters keys"
        return v


class TabularParameterProvider(EpisodeParameterProvider):
    """EpisodeParameterProvider that reads data from a tabular structure."""

    def __init__(self, **kwargs) -> None:
        self.config: TabularValidator
        super().__init__(**kwargs)

        assert all(isinstance(v, Parameter) for v in self.config.parameters.values())
        self._params = {k: OverridableParameterWrapper(v) for k, v in self.config.parameters.items()}

        self._index: int = 0

    def reset(self) -> None:
        self._index = 0

    @staticmethod
    def get_validator() -> type[TabularValidator]:
        """Get the validator for this class."""
        return TabularValidator

    def _do_get_params(self, rng: Randomness, env_epp_ctx: dict | None) -> tuple[ParameterModel, int | None, dict | None]:
        if env_epp_ctx:
            index = env_epp_ctx["index"]
        else:
            index = self._index
            self._index += 1
            env_epp_ctx = {"index": index}

        self._episode_id = self.config.data.index[index] if index < len(self.config.data) else None

        row = self.config.data.iloc[index % len(self.config.data)]

        pm = copy.deepcopy(self._params)
        for i in row.index:
            pm[tuple(i.split(self.config.separator))].override_value = row.loc[i]

        return pm, self._episode_id, env_epp_ctx



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Pong Game
This class implements pong and is independing of RL based on
https://github.com/techwithtim/Pong-Python
"""
from enum import Enum

import pygame
from pydantic import BaseModel, ConfigDict


class Colors(Enum):
    """
    Enumer ations of display colors
    """

    WHITE = (255, 255, 255)
    BLACK = (0, 0, 0)
    BLUE = (0, 0, 255)
    RED = (255, 0, 0)


class GameStatus(Enum):
    """
    Enumeration of game status
    """

    LEFT_WIN = 0
    RIGHT_WIN = 1
    IN_PROGRESS = 2


class Paddle(BaseModel):
    """
    paddle controlled by user
    """

    width: int = 20
    height: int = 100
    x: float = 0
    y: float = 0
    vel: float = 4
    color: tuple[int, int, int] = Colors.WHITE.value
    collision_on: bool = True
    health: int = 1
    current_health: int = 1
    ball_hits: int = 0

    def __init__(self, **kwarg):
        super().__init__(**kwarg)
        self.reset()
        self.current_health = self.health

    def move(self, up=True):
        """
        move paddle y position up or down by velocity value
        """
        if up:
            self.y -= self.vel
        else:
            self.y += self.vel

    def reset(self):
        """
        move paddle to initial position
        """
        self.current_health = self.health
        self.collision_on = True


class Ball(BaseModel):
    """
    Ball object
    """

    max_vel: int = 5
    radius: int = 7
    x: float = 0
    y: float = 0
    x_vel: float = 5
    y_vel: float = 0

    def move(self):
        """
        move ball using current velocity
        """
        self.x += self.x_vel
        self.y += self.y_vel


class PongRender:
    """
    Class to render the current pong game
    """

    def __init__(self, screen_width, screen_height):
        self.display_window = pygame.display.set_mode((screen_width, screen_height))
        self.score_font = pygame.font.SysFont("comicsans", 50)
        self.extra_info_font = pygame.font.SysFont("comicsans", 24)
        self.ball_color = Colors.WHITE.value
        self.left_paddle_color = Colors.BLUE.value
        self.right_paddle_color = Colors.RED.value

    def draw(self, pong, left_score, right_score, extra_info=None):
        """
        draw all objects and scores on display
        """

        self.display_window.fill(Colors.BLACK.value)
        left_score_text = self.score_font.render(f"{left_score}", 1, Colors.WHITE.value)
        right_score_text = self.score_font.render(f"{right_score}", 1, Colors.WHITE.value)
        self.display_window.blit(left_score_text, (self.display_window.get_width() // 4 - left_score_text.get_width() // 2, 20))
        self.display_window.blit(right_score_text, (self.display_window.get_width() * (3 / 4) - right_score_text.get_width() // 2, 20))

        if extra_info:
            up_down_offset = self.display_window.get_height() * 0.95
            for item in extra_info:
                extra_info_text = self.extra_info_font.render(item, 1, Colors.WHITE.value)
                text_left_right_pos = max(self.display_window.get_width() * 0.1 - extra_info_text.get_width() // 2, 0.05)
                text_up_down_pos = up_down_offset - extra_info_text.get_height()
                self.display_window.blit(extra_info_text, (text_left_right_pos, text_up_down_pos))
                up_down_offset = text_up_down_pos

        pygame.draw.rect(
            self.display_window,
            self.left_paddle_color,
            (pong.left_paddle.x, pong.left_paddle.y, pong.left_paddle.width, pong.left_paddle.height),
        )
        pygame.draw.rect(
            self.display_window,
            self.right_paddle_color,
            (pong.right_paddle.x, pong.right_paddle.y, pong.right_paddle.width, pong.right_paddle.height),
        )

        for i in range(10, self.display_window.get_height(), self.display_window.get_height() // 20):
            if i % 2 == 1:
                continue
            pygame.draw.rect(
                self.display_window,
                Colors.WHITE.value,
                (self.display_window.get_width() // 2 - 5, i, 10, self.display_window.get_height() // 20),
            )

        pygame.draw.circle(self.display_window, self.ball_color, (pong.ball.x, pong.ball.y), pong.ball.radius)

        pygame.display.update()

    def draw_win(self, win_text):
        """
        draw win text
        """
        text = self.score_font.render(win_text, 1, Colors.WHITE.value)
        self.display_window.blit(
            text,
            (self.display_window.get_width() // 2 - text.get_width() // 2, self.display_window.get_height() // 2 - text.get_height() // 2),
        )

        pygame.display.update()
        pygame.time.delay(5000)


class Pong(BaseModel):
    """
    Pong game
    """

    screen_width: int = 700
    screen_height: int = 500
    left_paddle: Paddle = Paddle()
    right_paddle: Paddle = Paddle()
    ball: Ball = Ball()
    health_enabled: bool = False
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init__(self, **kwarg):
        super().__init__(**kwarg)
        self.reset()

    def reset(self):
        """
        reset game
        """
        self.left_paddle.x = 10
        self.left_paddle.y = self.screen_height // 2 - self.left_paddle.height // 2
        self.right_paddle.x = self.screen_width - 10 - self.right_paddle.width
        self.right_paddle.y = self.screen_height // 2 - self.right_paddle.height // 2
        self.ball.x = self.screen_width // 2
        self.ball.y = self.screen_height // 2

    def handle_paddle_movement(self, keys, left_paddle, right_paddle):
        """Move paddle positions based on keyboard values

        Parameters
        ----------
        keys : list
            list of pygame key presses
        left_paddle : paddle
            left paddle in game
        right_paddle : Ball
            right pad  in game
        """

        if pygame.K_w in keys and left_paddle.y - left_paddle.vel >= 0:
            left_paddle.move(up=True)
        if pygame.K_s in keys and left_paddle.y + left_paddle.vel + left_paddle.height <= self.screen_height:
            left_paddle.move(up=False)

        if pygame.K_UP in keys and right_paddle.y - right_paddle.vel >= 0:
            right_paddle.move(up=True)
        if pygame.K_DOWN in keys and right_paddle.y + right_paddle.vel + right_paddle.height <= self.screen_height:
            right_paddle.move(up=False)

    def handle_collision(self, ball, left_paddle, right_paddle):
        """Update ball velocity based on ball and paddle positions

        Parameters
        ----------
        ball : Ball
            ball in pong game
        left_paddle : paddle
            left paddle in game
        right_paddle : Ball
            right pad  in game
        """

        if ball.y + ball.radius >= self.screen_height or ball.y - ball.radius <= 0:
            ball.y_vel *= -1

        if ball.x_vel < 0:
            if (
                left_paddle.collision_on
                and ball.y >= left_paddle.y
                and ball.y <= left_paddle.y + left_paddle.height
                and ball.x - ball.radius <= left_paddle.x + left_paddle.width
            ):
                ball.x_vel *= -1

                middle_y = left_paddle.y + left_paddle.height / 2
                difference_in_y = middle_y - ball.y
                reduction_factor = (left_paddle.height / 2) / ball.max_vel
                y_vel = difference_in_y / reduction_factor
                ball.y_vel = -1 * y_vel
                if self.health_enabled:
                    left_paddle.current_health -= 1
                    if left_paddle.current_health <= 0:
                        left_paddle.collision_on = False
                # Updates the ball hit count for the left paddle
                left_paddle.ball_hits += 1

        elif (
            right_paddle.collision_on
            and ball.y >= right_paddle.y
            and ball.y <= right_paddle.y + right_paddle.height
            and ball.x + ball.radius >= right_paddle.x
        ):
            ball.x_vel *= -1

            middle_y = right_paddle.y + right_paddle.height / 2
            difference_in_y = middle_y - ball.y
            reduction_factor = (right_paddle.height / 2) / ball.max_vel
            y_vel = difference_in_y / reduction_factor
            ball.y_vel = -1 * y_vel
            if self.health_enabled:
                right_paddle.current_health -= 1
                if right_paddle.current_health <= 0:
                    right_paddle.collision_on = False
            # Updates the ball hit count for the right paddle
            right_paddle.ball_hits += 1

    def step(self, keys: list) -> GameStatus:
        """Advance pong game

        Parameters
        ----------
        keys : list
            list of keys pressed on keyboard.

        Returns
        -------
        GameStatus
            status of game after step
        """

        game_status = GameStatus.IN_PROGRESS

        self.handle_paddle_movement(keys, self.left_paddle, self.right_paddle)

        self.ball.move()

        self.handle_collision(self.ball, self.left_paddle, self.right_paddle)

        if self.ball.x < 0:
            game_status = GameStatus.RIGHT_WIN
        elif self.ball.x > self.screen_width:
            game_status = GameStatus.LEFT_WIN

        return game_status


def main():
    """
    Interactive game
    """
    pygame.init()

    run = True
    clock = pygame.time.Clock()
    fps = 60

    left_score = 0
    right_score = 0
    winning_score = 1

    pong = Pong(screen_width=700, screen_height=500, health_enabled=True)
    pong_render = PongRender(pong.screen_width, pong.screen_height)

    while run:
        clock.tick(fps)
        pong_render.draw(pong, left_score, right_score)

        keys = pygame.key.get_pressed()

        pressed_keys = [key for key in [pygame.K_w, pygame.K_s, pygame.K_UP, pygame.K_DOWN] if keys[key]]

        game_status = pong.step(pressed_keys)

        if game_status is GameStatus.RIGHT_WIN:
            right_score += 1
            pong.reset()
        elif game_status is GameStatus.LEFT_WIN:
            left_score += 1
            pong.reset()

        won = False
        if left_score >= winning_score:
            won = True
            win_text = "Left Player Won!"
        elif right_score >= winning_score:
            won = True
            win_text = "Right Player Won!"

        if won:
            pong_render.draw_win(win_text)
            pong.reset()
            left_score = 0
            right_score = 0

        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                run = False
                break

    pygame.quit()


if __name__ == "__main__":
    main()



"""
-------------------------------------------------------------------------------
The Autonomous Capabilities Team (ACT3) Deep Reinforcement Learning (D-RL) Environment

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
-------------------------------------------------------------------------------

NaN check module
"""
import traceback

import numpy as np

from corl.libraries.units import Quantity


def print_trace():
    """
    The following function adds the trace back for the `NaN` check below.
    Ensures that the user knows which calling function caused the NaN
    - Typical that NaN is caused by damaged/broken platform
    """
    for line in traceback.format_stack():
        print(line.strip())


def recursive_nan_check(data, skip_trace=False):
    """
    does the nan check after quantities are extracted
    adds a stack trace for the nan checks should something fail the check
    """
    if isinstance(data, Quantity):
        recursive_nan_check(data.m, skip_trace=skip_trace)
    elif np.isscalar(data):
        if np.isnan(data):
            if not skip_trace:
                print_trace()
            raise ValueError("Data contains nan")
    # special case for repeated space
    elif isinstance(data, dict):
        list(map(recursive_nan_check, data.values()))
    elif isinstance(data, list):
        list(map(recursive_nan_check, data))
    elif data is None:
        if not skip_trace:
            print_trace()
        raise ValueError("Data contains nan/None")
    elif data.shape == ():
        if np.isnan(data):
            if not skip_trace:
                print_trace()
            raise ValueError("Data contains nan")
    elif len(data.shape) == 1:
        if any(np.isnan(data)):
            if not skip_trace:
                print_trace()
            raise ValueError("Data contains nan")
    elif np.isnan(data).any():
        if not skip_trace:
            print_trace()
        raise ValueError("Data contains nan")


def nan_check_result(data, skip_trace=False):
    """
    Checks for nan in np array
    """
    tmp_data = data
    if not isinstance(tmp_data, list | dict | Quantity):
        raise RuntimeError(
            "ERROR: nan_check_result was given an outer data structure that was not a "
            "Quantity, Dict, or List, sensors and controllers must always return one of these types "
            f"data struct was {data}"
        )
    recursive_nan_check(tmp_data, skip_trace=skip_trace)
    return data



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

import typing
from abc import ABC, abstractmethod

from pydantic import BaseModel, ImportString, TypeAdapter, ValidationError, validate_call
from ray.rllib import BaseEnv
from ray.rllib.evaluation.episode_v2 import EpisodeV2
from ray.rllib.policy import Policy
from ray.rllib.utils.typing import PolicyID

from corl.libraries.factory import Factory
from corl.simulators.base_platform import BasePlatform
from corl.simulators.base_simulator_state import BaseSimulatorState


class AgentConfig(BaseModel):
    """
    platform_config: any configuration needed for the simulator to
                initialize this platform and configure it in the sim class
    parts_list: a list of tuples where the first element is come python class path
                    of a BasePart, and then the second element is a configuration dictionary for that part

    Arguments:
        BaseModel: [description]
    """

    platform_config: dict[str, typing.Any] | BaseModel
    parts_list: list[tuple[ImportString, dict[str, typing.Any]]]


class BaseSimulatorValidator(BaseModel):
    """
    worker_index: what worker this simulator class is running on < used for render
    vector_index: what vector index this simulator class is running on < used for render
    agent_configs: the mapping of agent names to a dict describing the platform
    frame_rate: the rate the simulator should run at (in Hz)
    """

    worker_index: int = 0
    vector_index: int = 0
    agent_configs: typing.Mapping[str, AgentConfig]
    frame_rate: float = 1.0


class BaseSimulatorResetValidator(BaseModel):
    """
    Validator to use to validate the reset input to a simulator class
    allows the simulator class to take EPP params and structure/validate them

    Note that all attributes in this validator need to survive being parsed by validation_helper_units_and_parameters.

    Subclasses can redefine `platforms` to make the `typing.Any` more restrictive.  It must remain a dictionary with keys named for
    the platforms in the simulation.

    agent_configs_reset: the mapping of agent names to a dict describing the platform to be used during sim reset
    """

    platforms: dict[str, typing.Any] = {}
    agent_configs_reset: typing.Mapping[str, AgentConfig] = {}


@validate_call
def validation_helper_units_and_parameters(value: dict[str, typing.Any]) -> dict[str, typing.Any]:
    """Recursively inspect a dictionary, converting Factory"""

    output: dict[str, typing.Any] = {}
    for k, v in value.items():
        try:
            factory = TypeAdapter(Factory).validate_python(v)
        except ValidationError:
            pass
        else:
            output[k] = factory.build()
            continue

        try:
            output[k] = validation_helper_units_and_parameters(v)
        except ValidationError:
            pass
        else:
            continue

        output[k] = v

    return output


class SimulatorDefaultCallbacks:
    """This class allows the simulator to add simulator specific operation into the environment callbacks.

    These methods are called at the end of corl.environment.default_env_rllib_callbacks.EnvironmentDefaultCallbacks methods of the same
    name.
    """

    def on_episode_start(
        self,
        *,
        worker,
        base_env: BaseEnv,
        policies: dict[PolicyID, Policy],
        episode: EpisodeV2,
        **kwargs,
    ) -> None:
        """Callback run on the rollout worker before each episode starts.

        Args:
            worker: Reference to the current rollout worker.
            base_env: BaseEnv running the episode. The underlying
                sub environment objects can be retrieved by calling
                `base_env.get_sub_environments()`.
            policies: Mapping of policy id to policy objects. In single
                agent mode there will only be a single "default" policy.
            episode: EpisodeV2 object which contains the episode's
                state. You can use the `episode.user_data` dict to store
                temporary data, and `episode.custom_metrics` to store custom
                metrics for the episode.
            kwargs: Forward compatibility placeholder.
        """

    def on_episode_step(
        self,
        *,
        worker,
        base_env: BaseEnv,
        policies: dict[PolicyID, Policy] | None = None,
        episode: EpisodeV2,
        **kwargs,
    ) -> None:
        """Runs on each episode step.

        Args:
            worker: Reference to the current rollout worker.
            base_env: BaseEnv running the episode. The underlying
                sub environment objects can be retrieved by calling
                `base_env.get_sub_environments()`.
            policies: Mapping of policy id to policy objects.
                In single agent mode there will only be a single
                "default_policy".
            episode: EpisodeV2 object which contains episode
                state. You can use the `episode.user_data` dict to store
                temporary data, and `episode.custom_metrics` to store custom
                metrics for the episode.
            kwargs: Forward compatibility placeholder.
        """

    def on_episode_end(self, *, worker, base_env: BaseEnv, policies: dict[PolicyID, Policy], episode, **kwargs) -> None:
        """
        Runs at the end of the episode.

        Parameters
        ----------
        worker: RolloutWorker
            Reference to the current rollout worker.
        base_env: BaseEnv
            BaseEnv running the episode. The underlying
            env object can be gotten by calling base_env.get_sub_environments().
        policies: dict
            Mapping of policy id to policy objects. In single
            agent mode there will only be a single "default" policy.
        episode: MultiAgentEpisode
            EpisodeV2 object which contains episode
            state. You can use the `episode.user_data` dict to store
            temporary data, and `episode.custom_metrics` to store custom
            metrics for the episode.
        """


class BaseSimulator(ABC):
    """
    BaseSimulator is responsible for initializing the platform objects for a simulation
    and knowing how to setup episodes based on input parameters from a parameter provider
    it is also responsible for reporting the simulation state at each timestep
    """

    def __init__(self, **kwargs) -> None:
        self.config = self.get_simulator_validator(**kwargs)
        self.callbacks = self.get_callbacks()

    @property
    def get_simulator_validator(self) -> type[BaseSimulatorValidator]:
        """
        returns the validator for the configuration options to the simulator
        the kwargs to this class are validated and put into a defined struct
        potentially raising based on invalid configurations

        Returns:
            BaseSimulatorValidator -- The validator to use for this simulation class
        """
        return BaseSimulatorValidator

    @property
    def get_reset_validator(self) -> type[BaseSimulatorResetValidator]:
        """
        returns the validator that can be used to validate episode parameters
        coming into the reset function from the environment class

        Returns:
            BaseSimulatorResetValidator -- The validator to use during resets
        """
        return BaseSimulatorResetValidator

    @property
    def frame_rate(self) -> float:
        """Return the frame rate (in Hz) this simulator will run at"""
        return self.config.frame_rate

    def shutdown(self):
        """Shutdown the simulator. This will cause all resources to be closed. After this is called,
        reset/step may no longer be called and the simulation is ready for destruction. This should be
        safe to be called more than once, though subsequent calls should have no effect.
        """

    @abstractmethod
    def reset(self, config: dict[str, typing.Any]) -> BaseSimulatorState:
        """
        reset resets the simulation and sets up a new episode

        Arguments:
            config {typing.Dict[str, typing.Any]} -- The parameters to
                    validate and use to setup this episode

        Returns:
            BaseSimulatorState -- The simulation state
        """

    @abstractmethod
    def step(self, platforms_to_action: set[str]) -> BaseSimulatorState:
        """
        advances the simulation platforms and returns the state

        Arguments:
            platforms_to_action {typing.Dict[str, typing.Any]} -- the List of platform names who the
                environment believes should perform a new action on this call to step

        Returns:
            BaseSimulatorState -- The state after the simulation updates
        """

    @property
    @abstractmethod
    def sim_time(self) -> float:
        """
        returns the time

        Returns:
            float - time
        """

    @property
    @abstractmethod
    def platforms(self) -> dict[str, BasePlatform]:
        """
        returns a dict of platforms in the simulation

        Returns:
            dict of platforms mapped platform_name -> platform
        """

    @abstractmethod
    def mark_episode_done(self, done_info: dict, episode_state: dict, metadata: dict | None = None):
        """
        Takes in the done_info specifying how the episode completed
        and does any book keeping around ending an episode

        Arguments:
            done_info {OrderedDict} -- The Dict describing which Done conditions ended an episode
            episode_state {OrderedDict} -- The episode state at the end of the simulation
            metadata {dict | None}  --  Additional metadata to be stored along with the episode
        """

    @abstractmethod
    def save_episode_information(self, dones, rewards, observations, observation_units):
        """
        provides a way to save information about the current episode
        based on the environment

        Arguments:
            dones: the current done info of the step
            rewards: the reward info for this step
            observations: the observations for this step
            observation_units: the measurement units of the observations for this step
        """

    @property
    def get_callbacks(self) -> type[SimulatorDefaultCallbacks]:
        """Return the simulator callback class."""
        return SimulatorDefaultCallbacks

    def render(self, state, mode="human"):
        """
        allows you to do something to render your simulation
        you are responsible for checking which worker/vector index you are on
        """

    def delete_platform(self, name):
        """
        provides a way to delete a platform from the simulation
        """



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
property definitions for sensors
"""
from typing import Annotated

import numpy as np
from pydantic import Field, StrictFloat, StrictStr

from corl.libraries.property import BoxProp, DiscreteProp


class TimeProp(BoxProp):
    """
    TimeProp defines time space
    """

    name: str = "time"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [0.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [np.inf]
    unit: StrictStr = "seconds"
    description: str = "The time of the sensor"


class NoOpProp(DiscreteProp):
    """
    NoOpProp defines null action
    """

    name: str = "NoOp"
    n: int = 1
    description: str = "No op prop actions from this space will not be sent to simulator"



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from abc import abstractmethod
from collections.abc import Callable
from typing import Generic, TypeVar

from pydantic import BaseModel
from ray.rllib.algorithms.algorithm import Algorithm
from ray.rllib.algorithms.callbacks import DefaultCallbacks

T = TypeVar("T")


class InferenceCallbackValidator(BaseModel):
    """Base validator for InferenceCallback"""


class InferenceCallback(DefaultCallbacks, Generic[T]):
    def __init__(self, legacy_callbacks_dict: dict[str, Callable] | None = None):
        super().__init__(legacy_callbacks_dict)  # type:  ignore[arg-type]

    @staticmethod
    def get_validator() -> type[InferenceCallbackValidator]:
        """Return validator"""
        return InferenceCallbackValidator

    def on_algorithm_init(self, *, algorithm: Algorithm, **kwargs) -> None:
        super().on_algorithm_init(algorithm=algorithm, kwargs=kwargs)

        self.config = self.get_validator()(**dict(algorithm.config).get(f"{type(self).__name__}", {}))

    @abstractmethod
    def process_message(self, message: T):
        ...



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import logging
from collections.abc import Sequence
from pathlib import Path
from typing import Any, cast

import jsonargparse

from corl.evaluation.connection.base_eval_connection import ConnectionValidator, EvalConfig
from corl.evaluation.launchers.evaluate_runner import EvalRunner
from corl.evaluation.runners.section_factories.engine.rllib.rllib_trainer import ray_context
from corl.libraries.config_file_watcher import LoggingSetup
from corl.libraries.context import add_context
from corl.parsers.yaml_loader import load_file

#################################################################################
# If you want to debug a configuration that already exists
# Remove comments from the following line and make sure `from_dict` is commented out
# Then run this file
#################################################################################
# from_file = ""
#################################################################################
# If you haven't made a yaml file for your setup, you can create it here
# Remove comments from the following definition and make sure `from_file` is commented out
#################################################################################

# from_yml = """
# teams:
#     team_participant_map:
#         blue:
#         -   platform_config: config/platforms/f16/sixdof_lcirst_rbr_point_mass.yml
#             agents:
#             -
#                 name: blue0
#                 agent_config: config/tasks/1v1/agents/wld_slew_lcirst_delta.yml
#                 policy_config: config/policies/framestacking.yml
#                 agent_loader:
#                     class_path: corl.evaluation.loader.policy_checkpoint.policy_checkpoint
#                     init_args:
#                         checkpoint_filename: /tmp/data/corl/agents/dummy_1v1/checkpoint-12
#                         policy_id: blue0
#         red:
#         -   platform_config: config/platforms/f16/sixdof_lcirst_rbr_point_mass.yml
#             agents:
#             -
#                 name: red0
#                 agent_config: config/tasks/1v1/agents/wld_slew_lcirst_delta.yml
#                 policy_config: config/policies/framestacking.yml
#                 agent_loader:
#                     class_path: corl.evaluation.loader.policy_checkpoint.PolicyCheckpoint
#                     init_args:
#                         checkpoint_filename: /tmp/data/corl/agents/dummy_1v1/checkpoint-12
#                         policy_id: blue0

#     # If platform names ever change from <side><idx> format this will need to be changed
#     participant_id_schema: "%team_name%%idx%"
# task:
#   TaskGenerator:
#     task_config_file: config/tasks/1v1/tasks/multi_agent_large_wld.yml
# test_cases:
#   pandas:
#     data: config/evaluation/test_cases/1v1_f16_16.yml
#     source_form: FILE_YAML_CONFIGURATION
#     randomize: False
# recorders:
#     -
#         class_path: corl.evaluation.recording.Folder
#         init_args:
#             dir: /tmp/data/corl/evaluation_1v1
#             append_timestamp: False
# """

######################################################
# ONLY MODIFY BELOW THIS LINE IF KNOW WHAT DOING
######################################################


def get_args(path: str | None = None, alternate_argv: Sequence[str] | None = None) -> tuple[jsonargparse.Namespace, jsonargparse.Namespace]:
    """
    This is a method that allows for collection of arguments for evaluation

    Parameters:
    ----------
    path: str
        An optional argument to a path to config file that needs to be parsed to collect evaluation run arguments.

    Returns:

    instantiated: jsonargparse.Namespace
        A jsonargparse namespace containing instantiated objects to use for evaluation.
    args: jsonargparse.Namespace
        A jsonargparse namespace containing the basic run arguments to use for evaluaion. Used by the pipeline.

    """

    parser = jsonargparse.ArgumentParser()
    parser.add_argument("--cfg", type=Path, help="path to a json/yml file containing the running arguments")
    # parser.add_argument("--connection_cfg", type=Path, help="path to a json/yml file containing the connection arguments")

    # Temporary directory
    parser.add_argument("tmpdir_base", type=Path, default="/tmp")  # noqa: S108

    parser.add_argument("--include_dashboard", action="store_true")

    parser.add_argument(
        "--log-level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="INFO",
        help="Log level. This will be overridden by --log-config if log config exists.",
    )

    parser.add_argument("--log-config", type=Path, default=Path.cwd() / "logging.yml", help="Optional path to yaml logging configuration.")

    args = parser.parse_args(args=alternate_argv) if path is None else parser.parse_path(path)

    instantiated = parser.instantiate_classes(args)

    return instantiated, args


def load_config(config_file: Path) -> dict[str, Any]:
    """Loads config from file"""
    return cast(dict[str, Any], load_file(config_filename=config_file))


def main(instantiated_args: jsonargparse.Namespace, config: dict[str, Any]):
    """Main function block to evaluate from a configuration

    Parameters:
    ----------
        instantiated_args: jsonargparse.Namespace
            Contains as a Namespace the instantiated objects needed to run evaluation.
    """
    if "experiment" not in config:
        config = {"experiment": config}

    connection = ConnectionValidator(**config).connection

    with add_context({"connection": connection}):
        kwargs = {"path": instantiated_args.cfg, "raw_config": config, **config}
        eval_schema = EvalConfig(**kwargs)

        with ray_context(local_mode=eval_schema.experiment.engine.rllib.debug_mode, include_dashboard=instantiated_args.include_dashboard):
            eval_runner = EvalRunner(tmpdir_base=instantiated_args.tmpdir_base, **{**config, "connection": connection})
            eval_runner(eval_schema)
            eval_runner.run()


def pre_main(alternate_argv: Sequence[str] | None = None):
    """
    calls gets current args and passes them to main
    """
    instantiated, args = get_args(alternate_argv=alternate_argv)
    LoggingSetup(default_path=str(args.log_config), default_level=logging._nameToLevel[args.log_level])  # noqa: SLF001
    cfg = load_config(instantiated.cfg)
    main(instantiated, cfg)


if __name__ == "__main__":
    pre_main()



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Configures a rllib trainer to run evaluation
"""
import copy
import dataclasses
import logging
from contextlib import contextmanager
from pathlib import Path
from typing import Annotated

import ray
from pydantic import AfterValidator, BaseModel, Field, ImportString
from pydantic.dataclasses import dataclass
from ray.rllib.algorithms import Algorithm
from ray.rllib.algorithms.callbacks import DefaultCallbacks, make_multi_callbacks
from ray.rllib.utils.typing import EnvType
from ray.tune.registry import get_trainable_cls

from corl.environment.multi_agent_env import ACT3MultiAgentEnv, ACT3MultiAgentEnvValidator
from corl.episode_parameter_providers.remote import RemoteEpisodeParameterProvider
from corl.evaluation.eval_logger_name import EVAL_LOGGER_NAME
from corl.evaluation.runners.section_factories.engine.rllib.default_evaluation_callbacks import DefaultEvaluationCallbacks
from corl.evaluation.runners.section_factories.engine.rllib.eval_config_updates import RllibConfigUpdate
from corl.evaluation.runners.section_factories.engine.rllib.interruptable_callback import InterruptableCallback
from corl.evaluation.runners.section_factories.plugins.environment_state_extractor import EnvironmentStateExtractor
from corl.evaluation.runners.section_factories.plugins.platform_serializer import PlatformSerializer
from corl.evaluation.runners.section_factories.plugins.plugins import Plugins
from corl.evaluation.runners.section_factories.task import Experiment
from corl.evaluation.runners.section_factories.test_cases.test_case_manager import TestCaseStrategy
from corl.libraries.algorithm_helper import reset_epp
from corl.libraries.factory import Factory


@contextmanager
def ray_context(*args, **kwargs):
    """Context manager to initialize and shutdown ray if not already initialized

    This function does not reinitialized an existing ray.  It takes any parameters accepted by
    ray.init() and passes them directly to ray.init().
    """
    if not ray.is_initialized():
        print("+" * 100)
        print("Initialize the ray context - ray.init")
        print("+" * 100)
        ray.init(*args, **kwargs)
        try:
            yield
        finally:
            print("+" * 100)
            print("Shutdown the ray context - ray.shutdown")
            print("+" * 100)
            ray.shutdown()
    else:
        yield


def callback_factory(serializer: PlatformSerializer, environment_state_extractor: EnvironmentStateExtractor | None = None):
    """Factory method to inject the platform serilization object

    Returns:
        _type_: Class that implements DefaultEvaluationCallbacks, but with default constructor
    """

    class ConcreteCallbacks(DefaultEvaluationCallbacks):
        """Class factory

        Args:
            DefaultEvaluationCallbacks (_type_): _description_
        """

        def platform_serializer(self):  # noqa: PLR6301
            """
            Method specifying the serialize function
            """
            return serializer

        def extract_environment_state(self, env_state):  # noqa: PLR6301
            if environment_state_extractor is None:
                return {}
            return copy.deepcopy(environment_state_extractor.extract(env_state))

    return ConcreteCallbacks


def check_subclass(v):
    if not issubclass(v, DefaultCallbacks):
        raise ValueError(f"Invalid type: {v}, must subclass {DefaultCallbacks}")
    return v


@dataclass
class RllibTrainer:
    """Configuration and logic to generate a Rllib trainer instance"""

    # Evaluation setup items
    callbacks: list[Annotated[ImportString, AfterValidator(check_subclass)]] = Field(default_factory=list)

    # Evaluation Core
    trainer_cls: str | type[Algorithm] = dataclasses.field(default="PPO")
    debug_mode: bool = dataclasses.field(default=False)
    workers: int = dataclasses.field(default=0)
    envs_per_worker: int | None = dataclasses.field(default=None)
    horizon: int | None = dataclasses.field(default=None)
    explore: bool = dataclasses.field(default=False)
    output_dir: Path | None = None

    env_cls: type[ACT3MultiAgentEnv] = ACT3MultiAgentEnv
    env_validator_cls: type[ACT3MultiAgentEnvValidator] = ACT3MultiAgentEnvValidator

    def generate(
        self,
        experiment: Experiment,
        test_case_manager: TestCaseStrategy | None = None,
        plugins: Plugins | None = None,
    ) -> Algorithm:
        """Generate the algorithm

        Arguments:
            experiment {Experiment} -- Experiment to load
            output_dir {str} -- directory to output to
            pandas_test_cases {pd.DataFrame} -- test cases

        Returns:
            Algorithm -- Resulting algorithm object
        """

        logger = logging.getLogger(EVAL_LOGGER_NAME)

        #############################################
        ## Manipulate the rllib_config for evaluation
        assert experiment.rllib_config is not None
        rllib_config = copy.deepcopy(experiment.rllib_config)
        rllib_config["env_config"] = experiment.env_config
        rllib_config["env"] = self.env_cls

        # default config manipulation
        default_eval_config_update = RllibConfigUpdate(
            output_dir=self.output_dir,
            test_case_manager=test_case_manager,
            workers=self.workers,
            envs_per_worker=self.envs_per_worker,
            explore=self.explore,
        )
        rllib_config = default_eval_config_update.update(rllib_config)
        if self.horizon is not None:
            rllib_config["horizon"] = self.horizon
            rllib_config["env_config"]["horizon"] = self.horizon
            rllib_config["train_batch_size"] = self.horizon
            rllib_config["sgd_minibatch_size"] = min(1000, int(self.horizon / 10))
            rllib_config["rollout_fragment_length"] = self.horizon

        # Any other config manipulation
        if plugins is not None:
            for cfg_update in plugins.eval_config_update:
                rllib_config = cfg_update.update(rllib_config)

        #############################################
        ## Setup environment epps
        if self.workers > 0:
            # raise NotImplementedError("worker > 0 has not been implemented/tested")
            logger.info("Performing distributed rollout with %s workers", self.workers)

            rllib_config["env_config"]["episode_parameter_provider"] = RemoteEpisodeParameterProvider.wrap_epp_factory(
                Factory(**rllib_config["env_config"]["episode_parameter_provider"], namespace="evaluation"),
                actor_name=ACT3MultiAgentEnv.episode_parameter_provider_name,
            )

            for agent_name, agent_configs in rllib_config["env_config"]["agents"].items():
                agent_configs.class_config.config["episode_parameter_provider"] = RemoteEpisodeParameterProvider.wrap_epp_factory(
                    Factory(**agent_configs.class_config.config["episode_parameter_provider"], namespace="evaluation"), agent_name
                )
        else:
            logger.info("Performing serial rollout")

        # This constructs all the EPPs once and uses them to configure all the environments with
        rllib_config["env_config"]["epp_registry"] = self.env_validator_cls(**rllib_config["env_config"]).epp_registry

        #############################################
        ## Setup environment callbacks
        callbacks_list: list[type[DefaultCallbacks]] = [InterruptableCallback]

        # Add the default platform serialization callback
        if plugins is not None:
            callbacks_list.append(callback_factory(plugins.platform_serialization, plugins.environment_state_extractor))

        # add all other callbacks given to this module
        callbacks_list.extend(iter(self.callbacks))
        rllib_config["callbacks"] = make_multi_callbacks(callbacks_list)
        rllib_config["disable_env_checking"] = True

        #############################################
        ## Get the algorithm type
        trainable_cls: type[Algorithm]
        if isinstance(self.trainer_cls, str):
            trainable_cls = get_trainable_cls(self.trainer_cls)
        elif issubclass(self.trainer_cls, Algorithm):
            trainable_cls = self.trainer_cls
        else:
            raise RuntimeError('"trainer_cls" field must either be a string or a `Algorithm` type')

        #############################################
        ## Wrap the algorithm type with one that will kill any remote epps on cleanup
        class AlgorithmWithProperCleanup(trainable_cls):  # type: ignore
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

            def cleanup(self):
                def shutdown_simulator(env: EnvType):
                    if isinstance(env, ACT3MultiAgentEnv):
                        env.simulator.shutdown()

                def shutdown_epps(env: EnvType):
                    if isinstance(env, ACT3MultiAgentEnv):
                        for epp in env.config.epp_registry.values():
                            epp.shutdown()

                def cleanup_env(env: EnvType) -> list:
                    shutdown_simulator(env)
                    shutdown_epps(env)
                    return []

                if self.workers is not None:
                    self.workers.foreach_env(cleanup_env)

                super().cleanup()

        algorithm = AlgorithmWithProperCleanup(config=rllib_config, env=self.env_cls)

        #############################################
        ## Reset the epps as they may be in a non-default state after Algorithm construction
        reset_epp(algorithm)

        return algorithm


class RllibConfig(BaseModel):
    """Wrapper for RllibTrainer"""

    rllib: RllibTrainer



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

from pydantic import validator

from corl.evaluation.launchers.base_eval import BaseEvaluator, BaseEvaluatorValidator, EvalConfig

from .algorithm_runner import AlgorithmRunner


class InferenceValidator(BaseEvaluatorValidator):
    class Config:
        arbitrary_types_allowed = True

    algorithm_runner: AlgorithmRunner = None  # type: ignore

    @validator("algorithm_runner", pre=True, always=True)
    def construct_algorithm_runner(cls, v):
        if isinstance(v, AlgorithmRunner):
            return v

        if isinstance(v, dict):
            return AlgorithmRunner(**v)

        return AlgorithmRunner()


class Inference(BaseEvaluator):
    def __init__(self, **kwargs):
        self._config: InferenceValidator  # type: ignore
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> type[InferenceValidator]:
        return InferenceValidator

    def reset(self):
        self._config.algorithm_runner.stop(continue_running=True)

    def stop(self):
        self._config.algorithm_runner.stop()

    def __call__(self, config: EvalConfig, **kwargs):
        self._config.algorithm_runner.reset(config.experiment.algorithm)
        yield from self._config.algorithm_runner.run()



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import math
from collections import OrderedDict
from functools import cached_property

import gymnasium
import numpy as np

from corl.glues.base_wrapper import BaseWrapperGlue, BaseWrapperGlueValidator
from corl.glues.common.observe_sensor import ObserveSensor
from corl.libraries.property import BoxProp, DictProp
from corl.libraries.units import corl_get_ureg


class TrigValuesValidator(BaseWrapperGlueValidator):
    """
    Validator for TrigValues glue
    """

    wrapped: ObserveSensor
    observation: str = "direct_observation"
    index: int = 0


class TrigValues(BaseWrapperGlue):
    """
    Glue class for an observation of sin & cos of an angle
    """

    class Fields:
        """
        field data
        """

        COS = "cos"
        SIN = "sin"

    def __init__(self, **kwargs) -> None:
        self.config: TrigValuesValidator
        super().__init__(**kwargs)

        self._uname = f"{self.config.name}"

    @staticmethod
    def get_validator() -> type[TrigValuesValidator]:
        return TrigValuesValidator

    def get_unique_name(self):
        """Provides a unique name of the glue to differentiate  it from other glues."""
        return self._uname

    @cached_property
    def observation_prop(self):
        return DictProp(
            spaces={
                self.Fields.COS: BoxProp(low=[-1.0], high=[1.0], dtype=np.dtype("float32"), unit="dimensionless"),
                self.Fields.SIN: BoxProp(low=[-1.0], high=[1.0], dtype=np.dtype("float32"), unit="dimensionless"),
            }
        )

    def get_observation(self, other_obs: OrderedDict, obs_space, obs_units) -> OrderedDict:
        """Constructs a portion of the observation space

        Arguments:
            platforms {List[BasePlatform]} -- List of current platforms.
                Guaranteed to be length of 1 and set to self.obs_agent

        Returns:
            OrderedDict -- Dictionary containing observation in <FIELD> key
        """
        d = OrderedDict()
        angle_rad = self.config.wrapped.get_observation(other_obs, obs_space, obs_units)[self.config.observation][self.config.index].m_as(
            "rad"
        )

        _Quantity = corl_get_ureg().Quantity
        d[self.Fields.COS] = _Quantity(np.asarray([math.cos(angle_rad)], dtype=np.float32), "dimensionless")
        d[self.Fields.SIN] = _Quantity(np.asarray([math.sin(angle_rad)], dtype=np.float32), "dimensionless")
        return d

    @cached_property
    def action_space(self) -> gymnasium.spaces.Space | None:
        """No Actions"""
        return None

    def apply_action(self, action, observation, action_space, obs_space, obs_units):  # noqa: PLR6301
        """No Actions"""
        return



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from __future__ import annotations

from pydantic import BaseModel

from corl.evaluation.metrics.generator import MetricGenerator
from corl.libraries.factory import Factory
from corl.libraries.functor import Functor


class ScenarioMetricGenerators(BaseModel):
    """Represents a set of metric generators for a scenario"""

    world: list[MetricGenerator]
    default_participant: list[MetricGenerator]
    specific_participant: dict[str, list[MetricGenerator]]

    @staticmethod
    def _instantiate_config(config):
        """Iterate through keys for a config for a Functor and build anything that itself has a factory"""

        # Iteratively go though config and see if there is anything to use Factory for
        for config_key in config:
            if config[config_key] is not None and "type" in config[config_key]:
                config[config_key] = Factory.resolve_factory(config[config_key], {})

        return config

    @staticmethod
    def from_dict(data: dict) -> ScenarioMetricGenerators:
        """Create an instance from a dictionary"""

        if "world" in data:
            world_metric_generator = []
            for metric in data["world"]:
                functor = Functor(**metric)
                world_metric_generator.append(functor.create_functor_object())

        default_agent_metric_generators: list[MetricGenerator] = []
        agent_metric_generators: dict[str, list[MetricGenerator]] = {}
        if "agent" in data:
            for key in data["agent"]:
                if key == "__default__":
                    for metric in data["agent"]["__default__"]:
                        # Iteratively go though config and see if there is anything to use Factory for
                        functor = Functor(**metric)
                        default_agent_metric_generators.append(functor.create_functor_object())
                else:
                    for metric in data["agent"][key]:
                        functor = Functor(**metric)
                        if key not in agent_metric_generators:
                            agent_metric_generators[key] = []
                        agent_metric_generators[key].append(functor.create_functor_object())

        return ScenarioMetricGenerators(
            world=world_metric_generator,
            default_participant=default_agent_metric_generators,
            specific_participant=agent_metric_generators,
        )



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from collections import OrderedDict

import gymnasium
from pydantic import field_validator

from corl.dones.done_func_base import DoneFuncBase, DoneFuncBaseValidator, DoneStatusCodes
from corl.libraries.property import BoxProp
from corl.libraries.units import Quantity, corl_get_ureg
from corl.simulators.base_simulator import BaseSimulatorState
from corl.simulators.common_platform_utils import get_platform_by_name, get_sensor_by_name


class SensorBoundsCheckDoneValidator(DoneFuncBaseValidator):
    """Initialize an SensorBoundsCheckDone

    Parameters
    ----------
    min_value : float
        The minimum allowed value of this sensor
    max_value : float
        The maximum allowed value of this sensor
    sensor_name : str
        The name of the sensor to check
    """

    min_value: Quantity
    max_value: Quantity
    sensor_name: str

    @field_validator("max_value")
    def min_max_consistent(cls, v, values):
        """Validate that the maximum is bigger than the minimum."""
        if v.units != values.data["min_value"].units:
            raise ValueError(f'Inconsistent units for min and max: {values.data["min_value"].units}, {v.units}')
        if v.m <= values.data["min_value"].m:
            raise ValueError(f'Minimum bound {values.data["min_value"].m} exceeds maximum bound {v.m}')
        return v


class SensorBoundsCheckDone(DoneFuncBase):
    """
    Checks to see if a specified state parameter is within bounds
    """

    # True means that clients should pass ValueWithUnits rather than evaluated value
    REQUIRED_UNITS = {"min_value": True, "max_value": True, "sensor_name": None}

    def __init__(self, **kwargs) -> None:
        self.config: SensorBoundsCheckDoneValidator
        super().__init__(**kwargs)
        if self.config.name == type(self).__name__:
            self.config.name = f"{self.config.sensor_name}_Done"

    @staticmethod
    def get_validator() -> type[SensorBoundsCheckDoneValidator]:
        """Returns the validator for this done condition"""
        return SensorBoundsCheckDoneValidator

    def __call__(
        self,
        observation: OrderedDict,
        action: OrderedDict,
        next_observation: OrderedDict,
        next_state: BaseSimulatorState,
        observation_space: gymnasium.Space,
        observation_units: OrderedDict,
    ) -> bool:
        # Find Target Platform
        platform = get_platform_by_name(next_state, self.platform, allow_invalid=True)

        # platform does not exist
        if platform is None:
            return False

        # Find target Sensor
        sensor = get_sensor_by_name(platform, self.config.sensor_name)
        if not isinstance(sensor.measurement_properties, BoxProp):
            raise TypeError(f"Can only do bounds checking on BoxProp, received {type(sensor.measurement_properties).__name__}")

        # Get measured value
        measurement = sensor.get_measurement()
        assert len(measurement.m)
        if len(measurement.m) != 1:
            raise ValueError("Sensor measurement has more than one element")
        measured_value = corl_get_ureg().Quantity(value=measurement.m[0], units=sensor.measurement_properties.get_units())
        converted_value = measured_value.to(str(self.config.min_value.u))

        # Determine if done
        done = converted_value.m < self.config.min_value.m or self.config.max_value.m < converted_value.m

        if done:
            next_state.episode_state[self.platform][self.name] = DoneStatusCodes.LOSE

        return bool(done)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from collections.abc import Iterator
from contextlib import contextmanager
from contextvars import ContextVar
from typing import Any

from pydantic import BaseModel

_context_stack = ContextVar[list[dict[str, Any]]]("_context_stack", default=[])


@contextmanager
def add_context(value: dict[str, Any]) -> Iterator[None]:
    ctx: list[dict[str, Any]] = []
    ctx.extend(_context_stack.get())
    ctx.append(value)

    token = _context_stack.set(ctx)
    try:
        yield
    finally:
        _context_stack.reset(token)


def get_current_context() -> dict[str, Any]:
    ctx: dict[str, Any] = {}
    for d in _context_stack.get():
        ctx.update(d)
    return ctx


class BaseModelWithContext(BaseModel):
    def __init__(__pydantic_self__, **data: Any) -> None:
        __pydantic_self__.__pydantic_validator__.validate_python(
            data,
            self_instance=__pydantic_self__,
            context=get_current_context(),
        )



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Functors are objects the that can be treated as though they are a function.
When to use functors?
- Functors are used when you want to hide/abstract the real implementation.
Let's say you want to call the different functions depending on the input
but you don't want the user code to make explicit calls to those different
functions. This is the ideal situation where functors can help.
- In this scenario, we can go for a functor which internally calls the most
suitable function depending on the input.
- Now if later, none of functions to be called increases, then it would be
just a simple change in the backend code without disturbing any of the user
code. Thus functors help in creating maintainable, decoupled and extendable
 codes.
"""


class EnvFuncBase:
    """base definition for env functions"""

    def reset(self):
        """Base reset function for items such as rewards and dones"""

    @property
    def name(self) -> str:
        """gets the name for the functor

        Returns
        -------
        str
            The name of the functor
        """
        return type(self).__name__



"""
---------------------------------------------------------------------------


Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Layout for visualizing evaluations generated from training.
"""
import pickle
from collections.abc import Callable

import streamlit as st
from pydantic import validate_call

from corl.evaluation.recording.folder import FolderRecord
from corl.evaluation.runners.section_factories.engine.rllib.episode_artifact_logging_callback import EVAL_CHECKPOINTS_SUBDIR
from corl.libraries.functor import Functor
from corl.visualization.network_explainability.env_policy_transforms import AgentCheckpoint
from corl.visualization.plotly_animator import FrameAxisBounds
from corl.visualization.streamlit_app.default_layout import DefaultStreamlitPage, episode_artifact_tables
from corl.visualization.streamlit_app.sections.trajectory_animation import trajectory_playback_section
from corl.visualization.streamlit_app.utils import (
    _check_required_subdirectories_policy_viz,
    _group_checkpoints_by_input_output,
    create_selection_grid_policy_is_agent,
    create_selection_grid_single_and_self_policy,
    generate_grad_based_feat_importance,
    generate_vf_residuals_dist,
    init_network_explainability_module,
    st_directory_picker,
    text_section_policy_network_input,
)


class StreamlitPageEvaluationWhileTraining(DefaultStreamlitPage):
    """Streamlit page specific to evaluation runs generated from training"""

    def __init__(self, file_loader: Callable = FolderRecord) -> None:
        super().__init__(file_loader)
        self.epoch_selection: str = ""

    def display_directory_picker(self):
        """Displays the directory picker"""
        self._root_path, self._experiment_dir = st_directory_picker(title="Directory Picker")
        if self._experiment_dir is not None and self._root_path.joinpath(self._experiment_dir, "trajectories").exists():
            self.full_experiment_path = self._root_path.joinpath(self._experiment_dir)
            self.sess_state.chosen_experiment_path = True
        else:
            st.warning(
                f""" **Warning: Invalid Selection** - `{self.full_experiment_path}`
            does not contain episode artifacts. The episode
            artifacts are expected in a subdirectory named "trajectories".
            """
            )

    def display_epoch_picker(self):
        """Displays the epoch picker: this is specific to evaluation runs generated from training"""
        # This prevents a reload of the trajectory list while its still being populated
        if self.sess_state.get("chosen_experiment_path"):
            eval_dir = self.full_experiment_path.joinpath("trajectories/")
            self.sess_state.epoch_names = [f.name for f in eval_dir.iterdir()]
            self.sess_state.epoch_names.sort()
            self.sess_state.already_loaded_epochs_dir = True
            epoch_selection = st.selectbox("Pick an Epoch to Inspect:", self.sess_state.epoch_names, key="epoch_")
            if epoch_selection is not None:
                self.epoch_selection = epoch_selection

    def display_load_episode_artifact_option(self):
        """Displays the load evaluation button and poopulates the episode artifacts dict
        and tables_dict"""
        if st.button("Load Evaluation"):
            self.sess_state.loaded_evaluation = True
            self.tables_dict, self.episode_artifacts_dict = episode_artifact_tables(
                self.full_experiment_path.joinpath("trajectories", self.epoch_selection), file_loader=self.file_loader
            )
            self._set_obs_units_dict()

    def _construct_agent_checkpoints(self, use_placeholder=False):
        """Creates a list of agent checkpoints."""

        agent_checkpoints = []
        # Use the epoch selection if it exists, otherwise create one
        epoch_selection = self.epoch_selection
        if use_placeholder:
            epochs = list(self.full_experiment_path.joinpath(EVAL_CHECKPOINTS_SUBDIR).glob("epoch*"))
            epochs.sort()
            epoch_selection = epochs[0]

        checkpoints = self.full_experiment_path.joinpath(EVAL_CHECKPOINTS_SUBDIR, epoch_selection, "policies").glob("*")
        for policy_dir in checkpoints:
            agent_name = policy_dir.parts[-1]
            agent_checkpoints.append(AgentCheckpoint(agent_name=agent_name, checkpoint_dir=policy_dir))
        return agent_checkpoints

    @validate_call
    def display_trajectory_animation_section(self, animator: Functor, frame_axis_bounds: FrameAxisBounds | None = None):
        """
        Displays the Trajectory Animation Section

        Parameters
        ----------
        animator: Functor
            A Functor defining the animator
        frame_axis_bounds: FrameAxisBounds, optional
            Sets the default frame axis bounds for the animation.
        """
        self.sess_state = trajectory_playback_section(
            animator=animator,
            sess_state=self.sess_state,
            episode_artifacts_dict=self.episode_artifacts_dict,
            frame_axis_bounds=frame_axis_bounds,
        )

    def display_policy_network_viz_section(self) -> None:  # noqa: PLR0915
        # #########################################################################
        # Policy Network Visualization
        st.markdown("## Policy Network Visualization")
        st.markdown(f"**Training Run**: {self.full_experiment_path}")
        st.markdown(
            """
            ### Note
            - The compute time scales linearly with the number of checkpoints.
            - When running Policy Network Visualization for an active training run, make sure
            all trajectories have been generated for the epoch chosen, otherwise the results will change
            as more trajectories are generated.
            """
        )
        _check_required_subdirectories_policy_viz(self.full_experiment_path)

        # Use placeholder - we just want to grab the names of the agents at this point.
        agent_checkpoints = self._construct_agent_checkpoints(use_placeholder=True)
        policy_input_map = text_section_policy_network_input(agent_checkpoints=agent_checkpoints)
        columns = st.columns(3)
        single_policy_or_self_play = "Single-Policy / Self-Play"
        policy_is_agent = "PolicyIsAgent"
        clicked = False
        # #########################################################################
        # Select: Number of checkpoints to examine
        with columns[0]:
            num_checkpoints = st.number_input(label="Number of Checkpoints to Examine", value=1, min_value=1, step=1)
        with columns[1]:
            policy_map = st.radio(label="Policy Map", options=[single_policy_or_self_play, policy_is_agent], horizontal=True)
        # #########################################################################
        # Selected: SinglePolicy and Self-Play force a main-policy
        if policy_map == single_policy_or_self_play:
            with columns[2]:
                main_policy_name = st.selectbox(label="Main Policy Name", options=list(policy_input_map.keys()))
            num_selections_clicked = st.button("Generate Selections")
            if num_selections_clicked or hasattr(self.sess_state, "generated_single_policy_selfplay_selections"):
                agent_checkpoint_selections = create_selection_grid_single_and_self_policy(
                    num_checkpoints, self.full_experiment_path, main_policy=main_policy_name
                )
                clicked = st.button("Done")
                if clicked:
                    self.sess_state.agent_checkpoints = agent_checkpoint_selections
                    self.sess_state.generated_policy_viz_eval_train = False
                self.sess_state.generated_single_policy_selfplay_selections = True

        # #########################################################################
        # Selected: PolicyIsAgent allows the user to select both an epoch and
        # policy to review
        if policy_map == policy_is_agent:
            num_selections_clicked = st.button("Generate Selections")
            if num_selections_clicked or hasattr(self.sess_state, "generated_policy_is_agent_selections"):
                agent_checkpoint_selections = create_selection_grid_policy_is_agent(num_checkpoints, self.full_experiment_path)

                clicked = st.button("Done")
                if clicked:
                    self.sess_state.agent_checkpoints = agent_checkpoint_selections
                    self.sess_state.generated_policy_viz_eval_train = False

                self.sess_state.generated_policy_is_agent_selections = True

        # ####################################################################
        # Generates the Features: Only run when a selection is made and
        # visualizations have not been generated
        if clicked and self.sess_state.get("agent_checkpoints") and not self.sess_state.get("generated_policy_vz_eval_train", False):
            # Loads the environment configuration from params.pkl
            with open(self.full_experiment_path.joinpath("params.pkl"), "rb") as file_obj:
                params = pickle.load(file_obj)  # noqa: S301

            # Dictionary to hold the explainability module for each selection made
            # where the key is the shortened checkpoint path
            explain_modules: dict[str, dict] = {"explainer_class": {}, "vf_residuals": {}}
            # A list of dictionaries where each dictionary stores the
            # checkpoint path, policy inputs and policy outputs
            input_outputs = []
            explain_cls = None
            # Iterate over selections
            for agent_selections_dict in self.sess_state.agent_checkpoints:
                # Grab the agent checkpoint and the associated set of trajectories for the selected epoch
                agent_checkpoint = agent_selections_dict["agent_checkpoint"]
                episode_artifact_files = list(
                    self.full_experiment_path.joinpath("trajectories", agent_selections_dict["epoch_selected"]).glob("*.pickle")
                )

                # Initializes the explainability module
                explain_cls = init_network_explainability_module(
                    agent_checkpoints=[agent_checkpoint],
                    env_config=params["env_config"],
                    verbose=False,
                    policy_network_input_map=policy_input_map,
                    episode_artifact_files=episode_artifact_files,
                )

                checkpoint_path = str(agent_checkpoint.checkpoint_dir)
                prefix = str(self.full_experiment_path.joinpath(EVAL_CHECKPOINTS_SUBDIR))

                checkpoint_path = checkpoint_path.replace(prefix, "")

                # Store visualization outputs
                explain_modules["explainer_class"].update(
                    {checkpoint_path: {"class": explain_cls, "agent_id": agent_checkpoint.agent_name}}
                )
                explain_modules["vf_residuals"].update({checkpoint_path: explain_cls.vf_residuals})

                net_input_names = explain_cls.network_input_names.copy()
                net_output_names = explain_cls.network_output_names.copy()
                net_input_names.sort()
                net_output_names.sort()
                policy_input_string = ".".join(net_input_names)
                policy_output_string = ".".join(net_output_names)
                input_outputs.append({"checkpoint": checkpoint_path, "input": policy_input_string, "output": policy_output_string})

            # Group by policy networks with the same input and outputs
            if explain_cls is not None:
                groups = _group_checkpoints_by_input_output(input_outputs)
                self.sess_state.feat_importance_figs = []
                for group in groups:
                    grouped_feats = {}
                    for checkpoint_key in group:
                        explain_cls = explain_modules["explainer_class"][checkpoint_key]["class"]
                        agent_name = explain_modules["explainer_class"][checkpoint_key]["agent_id"]
                        assert explain_cls is not None
                        feats_dict = explain_cls.get_agg_feat_importance_grad_based(agent_id=agent_name)
                        grouped_feats[checkpoint_key] = feats_dict[agent_name]
                    assert explain_cls is not None
                    self.sess_state.feat_importance_figs.append(
                        {"feat_importance_grad_based": explain_cls.render_agg_feat_importance_grad_based(feats_dict=grouped_feats)}
                    )

                # iterate through the value function residuals and change the agent identifier to the checkpoint path
                combined_vf_residuals = {}
                for checkpoint_path, vf_residuals_dict in explain_modules["vf_residuals"].items():
                    assert len(vf_residuals_dict) == 1
                    for vf_residuals in vf_residuals_dict.values():
                        combined_vf_residuals[checkpoint_path] = vf_residuals
                    # Since we pass in the vf_residuals, we can use any instance of explain_cls
                    assert explain_cls is not None
                    fig = explain_cls.render_value_function_residual_dist(vf_residuals=combined_vf_residuals)
                    self.sess_state.residual_dist = fig
                # Flip the flag so that it does not rerun
                self.sess_state.generated_policy_viz_eval_train = True

        # When feat_importance_figs is defined, plot the figures for each group
        if hasattr(self.sess_state, "feat_importance_figs"):
            for fig_dict in self.sess_state.feat_importance_figs:
                generate_grad_based_feat_importance(fig_dict["feat_importance_grad_based"])
        if hasattr(self.sess_state, "residual_dist"):
            generate_vf_residuals_dist(self.sess_state.residual_dist)



from collections.abc import Callable

from corl.libraries.units import UnitRegistryConfiguration, corl_set_ureg


def ureg_setup_func(programatic_defines: list[Callable] | None = None):
    if programatic_defines is None:
        programatic_defines = []
    ureg = UnitRegistryConfiguration(programatic_defines=programatic_defines).create_registry_from_args()
    corl_set_ureg(ureg)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

from abc import ABC, abstractmethod


class EnvironmentStateExtractor(ABC):
    """Abstract class to define how to extract information that needs to be saved from the larger environment state"""

    @abstractmethod
    def extract(self, state: dict) -> dict:
        """extract information from the given state"""
        raise RuntimeError("Must Instantiate this method")



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
state Dict - Leverage https://github.com/ramazanpolat/StateDict -
"""
import copy
from collections import OrderedDict

DICT_RESERVED_KEYS = vars(OrderedDict).keys()


class StateDict(OrderedDict):
    """[summary]

    Parameters
    ----------
    OrderedDict : [type]
        [description]
    """

    def __init__(self, *args, **kwargs):
        self._recurse = kwargs.pop("recurse", True)
        super().__init__(*args, **kwargs)
        if self._recurse:
            super().__init__(StateDict.recursive_attrdict(self))

    @staticmethod
    def recursive_attrdict(obj):
        """Walks a simple data structure, converting dictionary to StateDict.
        Supports lists, tuples, and dictionaries.
        """
        ret = obj
        if isinstance(obj, dict):
            if issubclass(type(obj), StateDict):
                for k, v in obj.items():
                    obj[k] = StateDict.recursive_attrdict(v)

                ret = obj
            else:
                ret = StateDict(
                    {str(k): StateDict.recursive_attrdict(v) for (k, v) in obj.items()},
                    recurse=False,
                )
        elif isinstance(obj, list):
            ret = [StateDict.recursive_attrdict(i) for i in obj]
        elif isinstance(obj, tuple):
            ret = tuple(StateDict.recursive_attrdict(i) for i in obj)

        return ret

    def __setattr__(self, name, value):
        if isinstance(name, str) and name[0] != "_":
            self.__setitem__(name, value)
        super().__setattr__(name, value)

    def __getattr__(self, name):
        _item = self.get(name)  # __getitem__(key) if key in self else None

        if name not in self and _item is None and isinstance(name, str) and name[0] != "_":
            # attempt to ignore builtins and private attributes
            raise KeyError(f"{name}")

        return _item

    def __deepcopy__(self, memo):
        return StateDict(copy.deepcopy(dict(self)), recurse=self._recurse)

    def __delattr__(self, name):
        self.__delitem__(name)
        super().__delattr__(name)

    def __dir__(self):
        return list(super().__dir__()) + [str(k) for k in self.keys()]

    def keys(self):  # noqa: PLR6301
        return OrderedDict(sorted(super().items())).keys()

    def values(self):  # noqa: PLR6301
        return OrderedDict(sorted(super().items())).values()

    def items(self):  # noqa: PLR6301
        return OrderedDict(sorted(super().items())).items()

    @staticmethod
    def merge(dl):
        """[summary]

        Parameters
        ----------
        dl : [type]
            [description]

        Returns
        -------
        [type]
            [description]
        """
        r = {}
        for d in dl:
            r = StateDict(sorted(StateDict.merge_two(r, d)))
        return StateDict(sorted(r.items()))

    @staticmethod
    def stack_values(dl):
        """[summary]

        Parameters
        ----------
        dl : [type]
            [description]

        Returns
        -------
        [type]
            [description]
        """
        ak = set().union(*dl)
        r = {k: [] for k in ak}
        for d in dl:
            for k, v in d.items():
                r[k].append(v)
        return r

    @staticmethod
    def merge_two(d1, d2, replace=True):
        """Merges two dicts

        Parameters
        ----------
        d1 : [type]
            [description]
        d2 : [type]
            [description]
        replace : bool, optional
            [description], by default True

        Yields
        ------
        [type]
            [description]

        Raises
        ------
        an
            [description]
        """
        for k in set(d1.keys()).union(d2.keys()):
            if k in d1 and k in d2:
                if isinstance(d1[k], dict) and isinstance(d2[k], dict):
                    yield (k, StateDict(sorted(StateDict.merge_two(d1[k], d2[k]))))
                elif replace:
                    # replace the value on the first with the value in the second
                    yield (k, d2[k])
                else:
                    # maybe raise an exception, but default is to proceed without whining
                    pass
            elif k in d1:
                yield (k, d1[k])
            else:
                yield (k, d2[k])

    def to_dict(self) -> dict:
        """Converts to a dictionary

        Returns
        -------
        dict
            the dict form
        """
        temp = copy.deepcopy(self)
        for k, v in temp.items():
            if isinstance(v, StateDict):
                temp[k] = v.to_dict()
        return temp



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Pickle Pong State Callback
"""

import copy
import os
import pickle
from pathlib import Path

from ray.rllib import BaseEnv
from ray.rllib.algorithms.callbacks import DefaultCallbacks
from ray.rllib.evaluation.episode_v2 import EpisodeV2


class PongPickleEnvState(DefaultCallbacks):
    """
    This callback saves the environment state every fixed number of episodes
    (`save_every_n_episodes`). The accumulated awards at each time step and
    all contents of `env.state` are stored.

    The outputs are saved in the environment config's output path.

    """

    save_every_n_episodes: int = 100
    episode_counter: int = 0
    game_status_key: str = "game_status"
    rewards_accumulator_key: str = "rewards_accumulator"
    env_state_key: str = "env_state"

    def on_episode_step(
        self,
        *,
        base_env: BaseEnv,
        episode: EpisodeV2,
        **_,
    ) -> None:
        """Saves the environment state and accumulated rewards into
        `episode.user_data[self.env_state_key]`
        """

        if self.episode_counter % self.save_every_n_episodes == 0:
            env = base_env.get_sub_environments()[episode.env_id]
            if self.env_state_key not in episode.user_data:
                episode.user_data[self.env_state_key] = []

            env_state = copy.deepcopy(env.state)

            # saves the entire environment state object as a dictionary
            # and the accumulated reward at the this step.
            env_state = dict(env_state)
            env_state.update({self.rewards_accumulator_key: dict(episode.user_data[self.rewards_accumulator_key])})
            episode.user_data[self.env_state_key].append(env_state)

    def on_episode_end(self, *, worker, episode: EpisodeV2, **_) -> None:
        """
        Pickles the data and writes it out
        """
        if isinstance(episode, Exception):
            self.episode_counter += 1
            return

        if self.env_state_key in episode.user_data:
            env_states = episode.user_data[self.env_state_key]
            game_status = env_states[-1].get(self.game_status_key, "")
            if hasattr(game_status, "name"):
                game_status = game_status.name

            ball_hits_total = sum(simulator.paddle.ball_hits for _, simulator in env_states[-1]["sim_platforms"].items())
            count_as_string = str(self.episode_counter).zfill(8)
            output_directory = Path(worker.env.config.output_path) / self.env_state_key
            os.makedirs(output_directory, exist_ok=True)
            output_file = output_directory / f"episode_{count_as_string}_len_{len(env_states)}_ballhits_{ball_hits_total}_{game_status}.pkl"
            with open(output_file, "wb") as f:
                pickle.dump(env_states, f)
            print(f"Saved episode to: {output_file}")

            del episode.user_data[self.env_state_key]

        self.episode_counter += 1



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

import abc
import typing
from functools import cached_property

from corl.libraries.units import Quantity
from corl.simulators.base_parts import BaseController, BasePlatformPartValidator, BaseSensor


class CommsValidator(BasePlatformPartValidator):
    """Validator for comms"""

    default_value: typing.Any | None = None


class Comms(BaseController, BaseSensor, abc.ABC):
    """Abstract platform part for storing/communicating data"""

    def __init__(self, parent_platform, config, property_class) -> None:
        self.config: CommsValidator
        super().__init__(parent_platform, config, property_class)

        self._properties_config = str(config["properties"])
        self._last_measurement = self.default_value
        self.set_invalid()

    @staticmethod
    def get_validator() -> type[BasePlatformPartValidator]:
        """
        return the validator that will be used on the configuration
        of this part
        """
        return CommsValidator

    @cached_property
    def default_value(self) -> Quantity:
        """Return the default value for this part"""
        if self.config.default_value is not None:
            sample = self.config.default_value
            ret = self.measurement_properties.create_quantity(sample)
        else:
            space = self.measurement_properties
            ret = space.create_zero_sample()

        assert isinstance(ret, Quantity)
        return ret

    def _update_state(self, state):
        """Update state"""

    @abc.abstractmethod
    def _set_data(self, data: Quantity) -> None:
        """Write data to this part"""

    @abc.abstractmethod
    def _get_data(self) -> Quantity:
        """Read data from this part"""

    def apply_control(self, control) -> None:
        """
        The generic method to apply the control for this controller.

        Parameters
        ----------
        control
            The control to be executed by the controller
        """
        self.set_valid()
        assert isinstance(control, Quantity)
        self._set_data(control)

    def get_applied_control(self) -> Quantity:
        """
        Get the previously applied control that was given to the apply_control function
        Returns
        -------
        previously applied control that was given to the apply_control function
        """
        return self._get_data()

    def _calculate_measurement(self, state) -> Quantity:
        self._update_state(state)
        return self.get_applied_control()

    def calculate_and_cache_measurement(self, state):  # noqa: PLR6301
        """
        Calculates the measurement and caches the result in the _last_measurement variable

        Parameters
        ----------
        state: BaseSimulatorState
            The current state of the environment used to obtain the measurement
        """
        super().calculate_and_cache_measurement(state)


class TeamCommsValidator(CommsValidator):
    """Validator for team comms"""


class TeamComms(Comms):
    """Abstract platform part for communicating data between agents on the same team"""

    def __init__(self, parent_platform, config, property_class) -> None:
        self.config: TeamCommsValidator
        super().__init__(parent_platform, config, property_class)
        self.team = self.parent_platform.name[: [x.isdigit() for x in self.parent_platform.name].index(True)]

    @staticmethod
    def get_validator() -> type[BasePlatformPartValidator]:
        """
        return the validator that will be used on the configuration
        of this part
        """
        return TeamCommsValidator



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from pathlib import Path
from typing import Annotated

from fastapi import Request, status
from fastapi.responses import JSONResponse
from pydantic import BaseModel, BeforeValidator, ValidationError

from corl.evaluation.connection.base_eval_connection import BaseEvalConnection, BaseEvalConnectionValidator
from corl.evaluation.connection.base_rest_connection import BaseRestConnection, RestConnectionValidator
from corl.evaluation.launchers.base_eval import EvalConfig
from corl.evaluation.runners.inference.protocol.simple_epp_update import SimpleEppUpdate
from corl.libraries.context import add_context
from corl.parsers.yaml_loader import load_file


def build_config(v, info):
    if (v is None or len(v) == 0) and (path := info.data.get("path")):
        return load_file(config_filename=path)
    return v


class _ResetSchema(BaseModel):
    path: Path | None = None
    config: Annotated[dict | None, BeforeValidator(build_config)] = None


class RestEvalConnectionValidator(RestConnectionValidator, BaseEvalConnectionValidator):
    ...


class RestEvalConnection(BaseRestConnection, BaseEvalConnection[SimpleEppUpdate]):
    def __init__(self, **kwargs):
        external_delay_start = kwargs.get("delay_start", False)
        kwargs["delay_start"] = True
        super().__init__(**kwargs)

        self.app.post("/reset/")(self.reset_config)

        self.app.post("/epp/")(self.modify_epp)

        self.app.exception_handler(ValidationError)(self.validation_error_handler)

        if not external_delay_start:
            self.start()

    @staticmethod
    def get_validator() -> type[RestEvalConnectionValidator]:
        return RestEvalConnectionValidator

    def reset_config(self, data: _ResetSchema):
        with add_context({"connection": self}):
            eval_schema = EvalConfig(path=data.path, raw_config=data.config)
            self.reset_signal(eval_schema)

        return {"message": "ok"}

    def modify_epp(self, data: SimpleEppUpdate):
        self.modify_epp_signal(data)

        return {"message": "ok"}

    def validation_error_handler(self, request: Request, exc: ValidationError):  # noqa: PLR6301
        return JSONResponse(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            content=exc.json(),
        )



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
This module defines an evaluation specific ParameterProvider, which wraps a given ParameterProvider
and delegates parameter value generation to it. It increments an internal 'index' attribute up to
the provided 'num_test_cases' int, to provide an accurate episode_id during evaluation.

Author: John McCarroll
"""


from pydantic import ImportString

from corl.episode_parameter_providers import EpisodeParameterProvider, EpisodeParameterProviderValidator, ParameterModel, Randomness


class IncrementalWrapperValidator(EpisodeParameterProviderValidator):
    """
    Validation model for the inputs of IncrementalParameterProviderWrapper
    """

    num_test_cases: int
    type: ImportString  # noqa: A003 # the class of the wrapped ParameterProvider
    config: dict | None = {}  # the config for the wrapped ParameterProvider


class IncrementalParameterProviderWrapper(EpisodeParameterProvider):
    """
    A simple ParameterProvider wrapper, which delegates parameter value generation to the ParameterProvider it maintains.
    The added feature is an incremented 'index' attribute, used to track episode_id for the Evaluation framework.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.index = 0
        config = self.config.config if self.config.config is not None else {}  # type: ignore
        self.epp = self.config.type(type=self.config.type, parameters=self.config.parameters, **config)  # type: ignore

    def reset(self) -> None:
        self.index = 0

    @staticmethod
    def get_validator() -> type[IncrementalWrapperValidator]:
        """Get the validator for this class."""
        return IncrementalWrapperValidator

    def _do_get_params(self, rng: Randomness, env_epp_ctx: dict | None) -> tuple[ParameterModel, int | None, dict | None]:
        """
        This method retrieves parameter values from its wrapped ParameterProvider.
        It then returns them, alongside the episode_id.
        """

        # delegate to get params
        params, _, env_epp_ctx = self.epp.get_params(rng, env_epp_ctx)

        # get episode_id
        episode_id = self.index if self.index < self.config.num_test_cases else None  # type: ignore

        # increment after params collected
        self.index += 1

        return params, episode_id, env_epp_ctx

    # TODO: delegate other methods to wrapped EPP?



"""
This module defines an enumeration of available platform types that can be referenced in the platform config.
See config/tasks/docking_1d/docking1d_platform.yml.
"""

from __future__ import annotations

from corl.libraries.plugin_library import PluginLibrary
from corl.simulators.base_platform_type import BasePlatformType
from corl.simulators.docking_1d.simulator import Docking1dSimulator


class Docking1dAvailablePlatformType(BasePlatformType):
    """_summary_

    Args:
        PlatformType (_type_): _description_

    Raises:
        RuntimeError: _description_

    Returns:
        _type_: _description_
    """

    @classmethod
    def match_model(cls, config: dict) -> bool:
        if config["name"] == "DOCKING1D":
            return True
        return False


PluginLibrary.add_platform_to_sim(Docking1dAvailablePlatformType, Docking1dSimulator)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import dataclasses

import jsonargparse

from corl.evaluation.recording.i_recorder import IRecord


@jsonargparse.typing.final
@dataclasses.dataclass
class EvaluationArtifact_EvaluationOutcome:
    """
    Contains information needed to retrieve the evaluation outcome

    Args:
        location: Location of Evaluation outcome
    """

    location: IRecord


@jsonargparse.typing.final
@dataclasses.dataclass
class EvaluationArtifact_Metrics:
    """
    Contains information to describe artifacts from generating metrics
    Artifacts of metrics is a file stored on some location of the filesystem
    Later this may be extended to be connection to a database

    Args:
        location: Location of metrics artifact, currently only file system is supported
        metrics_file: Name of file to save metrics to. defaults to metrics.pkl
    """

    location: str  # Eventually this may be a reference to a database instead of just a path
    file: str = "metrics.pkl"


@jsonargparse.typing.final
@dataclasses.dataclass
class EvaluationArtifact_Visualization:
    """
    Contains information to describe visualization artifacts
    Artifacts of visualization are various files (depends on specific technique) stored on the filesystem
    Later this may be extended to be connection to a database

    Args:
        location: Location of plot artifacts
    """

    location: str  # Eventually this may be a reference to a database instead of just a path



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

from collections import OrderedDict, abc
from functools import cached_property

import numpy as np

from corl.glues.base_glue import BaseAgentPlatformGlue, BaseAgentPlatformGlueValidator
from corl.libraries.property import DictProp
from corl.libraries.units import Quantity
from corl.simulators.common_platform_utils import get_sensor_by_name


class ObserveSensorValidator(BaseAgentPlatformGlueValidator):
    """
    output_units: unit to convert the output data to
    sensor: which sensor to find on the platform
    """

    sensor: str
    output_units: str | None = None


class ObserveSensor(BaseAgentPlatformGlue):
    """Glue to observe a sensor

    Configuration:
        sensor: string      String of sensor class, sensor base name, or the PluginLibrary group name
                            For example consider a platform which has "SimOrientationRateSensor" attached.
                            The following will all be equivalent:
                            - SimOrientationRateSensor
                            - BaseOrientationRateSensor
                            - Sensor_OrientationRate

                            Important: If multiple sensors are found an array will be returned (ex. Gload)

                            This will also be the name attached to the glue, which can then be used
    """

    class Fields:
        """
        Fields in this glue
        """

        DIRECT_OBSERVATION = "direct_observation"

    @staticmethod
    def get_validator() -> type[ObserveSensorValidator]:
        return ObserveSensorValidator

    def __init__(self, **kwargs) -> None:
        self.config: ObserveSensorValidator
        super().__init__(**kwargs)

        self._sensor = get_sensor_by_name(self._platform, self.config.sensor)
        self._sensor_name: str = self.config.sensor

    def get_unique_name(self):
        """Class method that retrieves the unique name for the glue instance"""
        # TODO: This may result in ObserveSensor_Sensor_X, which is kinda ugly
        if self.config.name not in (ObserveSensor.__name__, "sensor"):
            return self.config.name
        return f"Obs_{self._sensor_name}"

    # TODO: broken
    def invalid_value(self) -> OrderedDict:
        """Return zeros when invalid"""
        arr: list[float] = []
        if isinstance(self._sensor.measurement_properties.low, abc.Sequence):  # type: ignore
            arr.extend([0.0 for _ in self._sensor.measurement_properties.low])  # type: ignore
        elif np.isscalar(self._sensor.measurement_properties.low):  # type: ignore
            arr.append(0.0)
        else:
            raise RuntimeError("Expecting either array or scalar")

        d = OrderedDict()
        d[self.Fields.DIRECT_OBSERVATION] = np.asarray(arr)
        return d

    @cached_property
    def observation_prop(self):
        tmp = self._sensor.measurement_properties
        if self.config.output_units:
            tmp = tmp.create_unit_converted_prop(self.config.output_units)
        return DictProp(spaces={self.Fields.DIRECT_OBSERVATION: tmp})

    def get_observation(self, other_obs: OrderedDict, obs_space, obs_units) -> OrderedDict:
        """Observation Values"""
        d: OrderedDict[str, Quantity] = OrderedDict()
        tmp = self._sensor.get_measurement()
        if self.config.output_units:
            assert isinstance(self.config.output_units, str)
            tmp = tmp.to(self.config.output_units)
        d[self.Fields.DIRECT_OBSERVATION] = tmp
        return d



"""
-------------------------------------------------------------------------------
The Autonomous Capabilities Team (ACT3) Deep Reinforcement Learning (D-RL) Environment

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
-------------------------------------------------------------------------------

PluginLibrary.AddClassToGroup(CLASS, "NAME", {"simulator": INTEGRATION_CLASS, "platform_type": AvailablePlatformTypes.XXXX})
"""
import importlib
import inspect
import itertools
import pkgutil
import sys
import typing
from traceback import print_tb

import numpy as np


class _PluginLibrary:
    """Class defining the plugin Library"""

    def __init__(self) -> None:
        self._groups: dict[str, list[tuple[typing.Callable, dict]]] = {}
        self.__PLATFORM_TYPES_LOC = "__PLATFORM_TYPES__"

    @staticmethod
    def add_paths(plugin_packages: list[str]):
        """
        loops through a list of strings (which are strings python paths)
        then recursively walks through subdirectories of those paths
        and imports them, which will cause any side effect of importing them
        such as adding a class to the plugin library
        """

        def pkg_error(module_name):
            print(f"Error in walk_packages module {module_name}")
            _, _, traceback = sys.exc_info()
            print_tb(traceback)
            raise ImportError

        for root_pkg in plugin_packages:
            root_import = importlib.import_module(root_pkg)
            for module in pkgutil.walk_packages(root_import.__path__, f"{root_import.__name__}.", onerror=pkg_error):
                importlib.import_module(module.name)

    def AddClassToGroup(self, regclass: typing.Callable, group_name: str, conditions: dict[str, list[typing.Any] | typing.Any]):
        """Add a type (or set of types) to a group name to be invoked with the provided conditions"""

        # TODO: delete this? I think? DO NOT MERGE CORL3 with out deleting or uncommenting
        # if there exists an entry with given group_name and conditions throw a RuntimeError
        # if group_name in self._groups:
        #     for group_to_check in self._groups[group_name]:
        #         if group_to_check[1] == conditions:
        #             raise RuntimeError(f"An instance with provided conditions has already been added: {group_name} -> {conditions}")

        conditions_list = []  # type: ignore
        if not isinstance(conditions, dict):
            raise RuntimeError(
                f"The conditions provided to register the class {regclass.__name__} "
                f"to group {group_name} were not a dictionary, were {type(conditions)}"
            )
        # if a user provides a list to a condition, this means they want to register more than one condition
        if any(isinstance(x, list) for x in conditions.values()):
            tmp_conditions = [cond if isinstance(cond, list) else [cond] for cond in conditions.values()]
            product_result = itertools.product(*tmp_conditions)

            conditions_list.extend(dict(zip(conditions.keys(), product_tuple)) for product_tuple in product_result)
        else:
            conditions_list.append(conditions)

        if not callable(regclass):
            raise RuntimeError(f"The {regclass} being registered to {group_name} is expected to be a callable. but it is {type(regclass)}")
        if inspect.isabstract(regclass):
            raise RuntimeError(
                f"The {regclass.__name__} being registered to {group_name} "
                f"is expected to be concrete. See MRO {inspect.getmro(type(regclass))}"
            )

        if group_name not in self._groups:
            self._groups[group_name] = []

        # TODO make sure the keys to provided conditions match with entries already in group
        for condition in conditions_list:
            self._groups[group_name].append((regclass, condition))

    def GroupExists(self, group_name: str) -> bool:
        """Determine if provided group name exists"""
        return group_name in self._groups

    def GroupMembers(self, group_name: str) -> list[tuple[typing.Callable, dict]]:
        """Return the members of the given group"""
        return self._groups[group_name]

    def FindGroup(self, reg_class: typing.Callable):
        """Return the group a class belongs to"""
        for group_name, group_list in self._groups.items():
            for item_tuple in group_list:
                if reg_class in item_tuple:
                    return group_name
        raise RuntimeError(f"Class {reg_class} not found in PluginLibrary")

    def FindMatch(self, group_name: str, condition: dict) -> typing.Callable:
        """Given a group and conditions find the associated list of types that match

        Returns None if the given group does not exists
        Raises an exception if the group exists but no match to provided could be identified
        """

        # first check that group_name exists in _group
        if group_name not in self._groups:
            raise RuntimeError(f"No items were found to be registered to group: {group_name}")

        tuple_items = self._groups[group_name]

        # in the case of no conditions
        if len(condition) == 0:
            if len(tuple_items) > 1:
                raise RuntimeError(f"In the group {group_name}, with no conditions, more than one match was established")
            return tuple_items[0][0]

        # otherwise determine the best match, or even if there is a match
        mapping_results = [difference_metric(condition, reg_tuple[1]) for reg_tuple in tuple_items]

        if np.allclose(mapping_results, 0):
            raise RuntimeError(f"In the group {group_name}, an instance matching the given conditions {condition} could not be established")

        max_index = np.argmax(mapping_results)

        return tuple_items[max_index][0]

    def add_platform_to_sim(self, platform_class: typing.Callable, sim_class: typing.Callable) -> None:
        """
        Utility for adding a platform mapping to a simulator
        """
        PluginLibrary.AddClassToGroup(platform_class, self.__PLATFORM_TYPES_LOC, {"platform": platform_class, "simulator": sim_class})

    def get_platform_from_sim_and_config(self, sim_class: typing.Callable, config: dict[str, typing.Any]) -> typing.Callable:
        """
        Utility for getting a platform type from the sim class and platform config dict
        """
        found_platforms = [
            env_plt[0]
            for env_plt in PluginLibrary.GroupMembers(self.__PLATFORM_TYPES_LOC)
            if env_plt[1]["simulator"] == sim_class and env_plt[0].match_model(config)  # type: ignore
        ]

        if not found_platforms:
            found_platforms = [
                env_plt[0] for env_plt in PluginLibrary.GroupMembers(self.__PLATFORM_TYPES_LOC) if env_plt[1]["simulator"] == sim_class
            ]
            raise RuntimeError(
                f"Unable to find a matching platform for the target configuration\n{config}"
                f"platforms classes registered to sim but not matching were: {found_platforms}"
            )

        if len(found_platforms) > 1:
            raise RuntimeError(
                f"Found {len(found_platforms)} matching platform for the target configuration\n{config}\n"
                f"platform classes found were: {found_platforms}"
            )

        return found_platforms[0]


def difference_metric(x1: dict[str, typing.Any], x2: dict[str, typing.Any]):
    """
    Heuristic for determining how much of a match 2 dictionaries are
    Dictionaries may only be a single layer or this may not work correctly

    This heuristic is calculated as follows

    number of conditions matches /
    (total number of keys in both dicts /2)

    this allows any match to get selected, and prioritize
    dictionaries with more matches

    in the event that x2 is empty the match score is set to
    0.5 / (total number of keys in both dicts /2). This essentially
    is a 0.5 match of a key and allows dictionaries with a single
    match to have a higher score while empty dicts to result in a
    non-zero match. Specifically, this is used for plugins with
    no conditions.

    Furthermore, all keys that exist in both x1 and x2 must match or a
    metric of 0 is returned.

    Arguments:
        x1 {dict} -- The dictionary used to query the list of dictionaries
        x2 {dict} -- The dictionary being compared to x1

    Returns:
        float -- The heuristic value of the comparison between the two
                 1 is max - every single field matched)
                 0 is minimum - no field matches)
                 0 is minimum - field present in both dictionaries does not match
    """

    match_score = 0
    total = (len(x1) + len(x2)) / 2

    if not x2:
        return 0.5 / total

    for key, value in x2.items():
        if key not in x1:
            continue
        if value == x1[key]:
            match_score += 1
        else:
            return 0
    return match_score / total


PluginLibrary = _PluginLibrary()



"""
This module defines an enumeration of available platform types that can be referenced in the platform config.
See config/tasks/docking_1d/docking1d_platform.yml.
"""

from __future__ import annotations

from corl.libraries.plugin_library import PluginLibrary
from corl.simulators.base_platform_type import BasePlatformType
from corl.simulators.pong.simulator import PongSimulator


class PongAvailablePlatformType(BasePlatformType):
    """_summary_

    Args:
        PlatformType (_type_): _description_

    Returns:
        _type_: _description_
    """

    @classmethod
    def match_model(cls, config: dict) -> bool:
        """_summary_

        Args:
            config (dict): _description_

        Returns:
            bool: _description_
        """
        if config["name"] == "PADDLE":
            return True
        return False


# BaseAvailablePlatformTypes children must be registered with the PluginLibrary, associated with a name, and their
# environment's Simulator class.

PluginLibrary.add_platform_to_sim(PongAvailablePlatformType, PongSimulator)
